# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""The fsl module provides classes for interfacing with the `FSL
<http://www.fmrib.ox.ac.uk/fsl/index.html>`_ command line tools.  This
was written to work with FSL version 4.1.4.
"""
import os
import os.path as op
from warnings import warn

import numpy as np
from nibabel import load

from ... import LooseVersion
from ...utils.filemanip import split_filename
from ..base import (
    TraitedSpec,
    File,
    InputMultiPath,
    OutputMultiPath,
    Undefined,
    traits,
    isdefined,
)
from .base import FSLCommand, FSLCommandInputSpec, Info


class BETInputSpec(FSLCommandInputSpec):
    # We use position args here as list indices - so a negative number
    # will put something on the end
    in_file = File(
        exists=True,
        desc="input file to skull strip",
        argstr="%s",
        position=0,
        mandatory=True,
        copyfile=False,
    )
    out_file = File(
        desc="name of output skull stripped image",
        argstr="%s",
        position=1,
        genfile=True,
        hash_files=False,
    )
    outline = traits.Bool(desc="create surface outline image", argstr="-o")
    mask = traits.Bool(desc="create binary mask image", argstr="-m")
    skull = traits.Bool(desc="create skull image", argstr="-s")
    no_output = traits.Bool(argstr="-n", desc="Don't generate segmented output")
    frac = traits.Float(desc="fractional intensity threshold", argstr="-f %.2f")
    vertical_gradient = traits.Float(
        argstr="-g %.2f",
        desc="vertical gradient in fractional intensity threshold (-1, 1)",
    )
    radius = traits.Int(argstr="-r %d", units="mm", desc="head radius")
    center = traits.List(
        traits.Int,
        desc="center of gravity in voxels",
        argstr="-c %s",
        minlen=0,
        maxlen=3,
        units="voxels",
    )
    threshold = traits.Bool(
        argstr="-t", desc="apply thresholding to segmented brain image and mask"
    )
    mesh = traits.Bool(argstr="-e", desc="generate a vtk mesh brain surface")
    # the remaining 'options' are more like modes (mutually exclusive) that
    # FSL actually implements in a shell script wrapper around the bet binary.
    # for some combinations of them in specific order a call would not fail,
    # but in general using more than one of the following is clearly not
    # supported
    _xor_inputs = (
        "functional",
        "reduce_bias",
        "robust",
        "padding",
        "remove_eyes",
        "surfaces",
        "t2_guided",
    )
    robust = traits.Bool(
        desc="robust brain centre estimation (iterates BET several times)",
        argstr="-R",
        xor=_xor_inputs,
    )
    padding = traits.Bool(
        desc=(
            "improve BET if FOV is very small in Z (by temporarily padding "
            "end slices)"
        ),
        argstr="-Z",
        xor=_xor_inputs,
    )
    remove_eyes = traits.Bool(
        desc="eye & optic nerve cleanup (can be useful in SIENA)",
        argstr="-S",
        xor=_xor_inputs,
    )
    surfaces = traits.Bool(
        desc=(
            "run bet2 and then betsurf to get additional skull and scalp "
            "surfaces (includes registrations)"
        ),
        argstr="-A",
        xor=_xor_inputs,
    )
    t2_guided = File(
        desc="as with creating surfaces, when also feeding in "
        "non-brain-extracted T2 (includes registrations)",
        argstr="-A2 %s",
        xor=_xor_inputs,
    )
    functional = traits.Bool(argstr="-F", xor=_xor_inputs, desc="apply to 4D fMRI data")
    reduce_bias = traits.Bool(
        argstr="-B", xor=_xor_inputs, desc="bias field and neck cleanup"
    )


class BETOutputSpec(TraitedSpec):
    out_file = File(desc="path/name of skullstripped file (if generated)")
    mask_file = File(desc="path/name of binary brain mask (if generated)")
    outline_file = File(desc="path/name of outline file (if generated)")
    meshfile = File(desc="path/name of vtk mesh file (if generated)")
    inskull_mask_file = File(desc="path/name of inskull mask (if generated)")
    inskull_mesh_file = File(desc="path/name of inskull mesh outline (if generated)")
    outskull_mask_file = File(desc="path/name of outskull mask (if generated)")
    outskull_mesh_file = File(desc="path/name of outskull mesh outline (if generated)")
    outskin_mask_file = File(desc="path/name of outskin mask (if generated)")
    outskin_mesh_file = File(desc="path/name of outskin mesh outline (if generated)")
    skull_mask_file = File(desc="path/name of skull mask (if generated)")
    skull_file = File(desc="path/name of skull file (if generated)")


class BET(FSLCommand):
    """FSL BET wrapper for skull stripping

    For complete details, see the `BET Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/BET/UserGuide>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> btr = fsl.BET()
    >>> btr.inputs.in_file = 'structural.nii'
    >>> btr.inputs.frac = 0.7
    >>> btr.inputs.out_file = 'brain_anat.nii'
    >>> btr.cmdline
    'bet structural.nii brain_anat.nii -f 0.70'
    >>> res = btr.run() # doctest: +SKIP

    """

    _cmd = "bet"
    input_spec = BETInputSpec
    output_spec = BETOutputSpec

    def _run_interface(self, runtime):
        # The returncode is meaningless in BET.  So check the output
        # in stderr and if it's set, then update the returncode
        # accordingly.
        runtime = super(BET, self)._run_interface(runtime)
        if runtime.stderr:
            self.raise_exception(runtime)
        return runtime

    def _format_arg(self, name, spec, value):
        formatted = super(BET, self)._format_arg(name, spec, value)
        if name == "in_file":
            # Convert to relative path to prevent BET failure
            # with long paths.
            return op.relpath(formatted, start=os.getcwd())
        return formatted

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        # Generate default output filename if non specified.
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix="_brain")
            # Convert to relative path to prevent BET failure
            # with long paths.
            return op.relpath(out_file, start=os.getcwd())
        return out_file

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = os.path.abspath(self._gen_outfilename())

        basename = os.path.basename(outputs["out_file"])
        cwd = os.path.dirname(outputs["out_file"])
        kwargs = {"basename": basename, "cwd": cwd}

        if (isdefined(self.inputs.mesh) and self.inputs.mesh) or (
            isdefined(self.inputs.surfaces) and self.inputs.surfaces
        ):
            outputs["meshfile"] = self._gen_fname(
                suffix="_mesh.vtk", change_ext=False, **kwargs
            )
        if (isdefined(self.inputs.mask) and self.inputs.mask) or (
            isdefined(self.inputs.reduce_bias) and self.inputs.reduce_bias
        ):
            outputs["mask_file"] = self._gen_fname(suffix="_mask", **kwargs)
        if isdefined(self.inputs.outline) and self.inputs.outline:
            outputs["outline_file"] = self._gen_fname(suffix="_overlay", **kwargs)
        if isdefined(self.inputs.surfaces) and self.inputs.surfaces:
            outputs["inskull_mask_file"] = self._gen_fname(
                suffix="_inskull_mask", **kwargs
            )
            outputs["inskull_mesh_file"] = self._gen_fname(
                suffix="_inskull_mesh", **kwargs
            )
            outputs["outskull_mask_file"] = self._gen_fname(
                suffix="_outskull_mask", **kwargs
            )
            outputs["outskull_mesh_file"] = self._gen_fname(
                suffix="_outskull_mesh", **kwargs
            )
            outputs["outskin_mask_file"] = self._gen_fname(
                suffix="_outskin_mask", **kwargs
            )
            outputs["outskin_mesh_file"] = self._gen_fname(
                suffix="_outskin_mesh", **kwargs
            )
            outputs["skull_mask_file"] = self._gen_fname(suffix="_skull_mask", **kwargs)
        if isdefined(self.inputs.skull) and self.inputs.skull:
            outputs["skull_file"] = self._gen_fname(suffix="_skull", **kwargs)
        if isdefined(self.inputs.no_output) and self.inputs.no_output:
            outputs["out_file"] = Undefined
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._gen_outfilename()
        return None


class FASTInputSpec(FSLCommandInputSpec):
    """Defines inputs (trait classes) for FAST"""

    in_files = InputMultiPath(
        File(exists=True),
        copyfile=False,
        desc="image, or multi-channel set of images, " "to be segmented",
        argstr="%s",
        position=-1,
        mandatory=True,
    )
    out_basename = File(desc="base name of output files", argstr="-o %s")
    # ^^ uses in_file name as basename if none given
    number_classes = traits.Range(
        low=1, high=10, argstr="-n %d", desc="number of tissue-type classes"
    )
    output_biasfield = traits.Bool(desc="output estimated bias field", argstr="-b")
    output_biascorrected = traits.Bool(
        desc="output restored image (bias-corrected image)", argstr="-B"
    )
    img_type = traits.Enum(
        (1, 2, 3),
        desc="int specifying type of image: (1 = T1, 2 = T2, 3 = PD)",
        argstr="-t %d",
    )
    bias_iters = traits.Range(
        low=1,
        high=10,
        argstr="-I %d",
        desc="number of main-loop iterations during " "bias-field removal",
    )
    bias_lowpass = traits.Range(
        low=4,
        high=40,
        desc="bias field smoothing extent (FWHM) " "in mm",
        argstr="-l %d",
        units="mm",
    )
    init_seg_smooth = traits.Range(
        low=0.0001,
        high=0.1,
        desc="initial segmentation spatial "
        "smoothness (during bias field "
        "estimation)",
        argstr="-f %.3f",
    )
    segments = traits.Bool(
        desc="outputs a separate binary image for each " "tissue type", argstr="-g"
    )
    init_transform = File(
        exists=True,
        desc="<standard2input.mat> initialise" " using priors",
        argstr="-a %s",
    )
    other_priors = InputMultiPath(
        File(exist=True),
        desc="alternative prior images",
        argstr="-A %s",
        minlen=3,
        maxlen=3,
    )
    no_pve = traits.Bool(
        desc="turn off PVE (partial volume estimation)", argstr="--nopve"
    )
    no_bias = traits.Bool(desc="do not remove bias field", argstr="-N")
    use_priors = traits.Bool(desc="use priors throughout", argstr="-P")
    # ^^ Must also set -a!, mutually inclusive?? No, conditional mandatory... need to figure out how to handle with traits.
    segment_iters = traits.Range(
        low=1,
        high=50,
        desc="number of segmentation-initialisation" " iterations",
        argstr="-W %d",
    )
    mixel_smooth = traits.Range(
        low=0.0, high=1.0, desc="spatial smoothness for mixeltype", argstr="-R %.2f"
    )
    iters_afterbias = traits.Range(
        low=1,
        high=20,
        desc="number of main-loop iterations " "after bias-field removal",
        argstr="-O %d",
    )
    hyper = traits.Range(
        low=0.0, high=1.0, desc="segmentation spatial smoothness", argstr="-H %.2f"
    )
    verbose = traits.Bool(desc="switch on diagnostic messages", argstr="-v")
    manual_seg = File(
        exists=True, desc="Filename containing intensities", argstr="-s %s"
    )
    probability_maps = traits.Bool(
        desc="outputs individual probability maps", argstr="-p"
    )


class FASTOutputSpec(TraitedSpec):
    """Specify possible outputs from FAST"""

    tissue_class_map = File(
        exists=True,
        desc="path/name of binary segmented volume file"
        " one val for each class  _seg",
    )
    tissue_class_files = OutputMultiPath(
        File(
            desc=(
                "path/name of binary segmented volumes one file for each class  "
                "_seg_x"
            )
        )
    )
    restored_image = OutputMultiPath(
        File(
            desc=(
                "restored images (one for each input image) named according to "
                "the input images _restore"
            )
        )
    )

    mixeltype = File(desc="path/name of mixeltype volume file _mixeltype")

    partial_volume_map = File(desc="path/name of partial volume file _pveseg")
    partial_volume_files = OutputMultiPath(
        File(desc="path/name of partial volumes files one for each class, _pve_x")
    )

    bias_field = OutputMultiPath(File(desc="Estimated bias field _bias"))
    probability_maps = OutputMultiPath(
        File(desc="filenames, one for each class, for each input, prob_x")
    )


class FAST(FSLCommand):
    """FSL FAST wrapper for segmentation and bias correction

    For complete details, see the `FAST Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FAST>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> fastr = fsl.FAST()
    >>> fastr.inputs.in_files = 'structural.nii'
    >>> fastr.inputs.out_basename = 'fast_'
    >>> fastr.cmdline
    'fast -o fast_ -S 1 structural.nii'
    >>> out = fastr.run()  # doctest: +SKIP

    """

    _cmd = "fast"
    input_spec = FASTInputSpec
    output_spec = FASTOutputSpec

    def _format_arg(self, name, spec, value):
        # first do what should be done in general
        formatted = super(FAST, self)._format_arg(name, spec, value)
        if name == "in_files":
            # FAST needs the -S parameter value to correspond to the number
            # of input images, otherwise it will ignore all but the first
            formatted = "-S %d %s" % (len(value), formatted)
        return formatted

    def _list_outputs(self):
        outputs = self.output_spec().get()
        if not isdefined(self.inputs.number_classes):
            nclasses = 3
        else:
            nclasses = self.inputs.number_classes
        # when using multichannel, results basename is based on last
        # input filename
        _gen_fname_opts = {}
        if isdefined(self.inputs.out_basename):
            _gen_fname_opts["basename"] = self.inputs.out_basename
            _gen_fname_opts["cwd"] = os.getcwd()
        else:
            _gen_fname_opts["basename"] = self.inputs.in_files[-1]
            _gen_fname_opts["cwd"], _, _ = split_filename(_gen_fname_opts["basename"])

        outputs["tissue_class_map"] = self._gen_fname(suffix="_seg", **_gen_fname_opts)
        if self.inputs.segments:
            outputs["tissue_class_files"] = []
            for i in range(nclasses):
                outputs["tissue_class_files"].append(
                    self._gen_fname(suffix="_seg_%d" % i, **_gen_fname_opts)
                )
        if isdefined(self.inputs.output_biascorrected):
            outputs["restored_image"] = []
            if len(self.inputs.in_files) > 1:
                # for multi-image segmentation there is one corrected image
                # per input
                for val, f in enumerate(self.inputs.in_files):
                    # image numbering is 1-based
                    outputs["restored_image"].append(
                        self._gen_fname(
                            suffix="_restore_%d" % (val + 1), **_gen_fname_opts
                        )
                    )
            else:
                # single image segmentation has unnumbered output image
                outputs["restored_image"].append(
                    self._gen_fname(suffix="_restore", **_gen_fname_opts)
                )

        outputs["mixeltype"] = self._gen_fname(suffix="_mixeltype", **_gen_fname_opts)
        if not self.inputs.no_pve:
            outputs["partial_volume_map"] = self._gen_fname(
                suffix="_pveseg", **_gen_fname_opts
            )
            outputs["partial_volume_files"] = []
            for i in range(nclasses):
                outputs["partial_volume_files"].append(
                    self._gen_fname(suffix="_pve_%d" % i, **_gen_fname_opts)
                )
        if self.inputs.output_biasfield:
            outputs["bias_field"] = []
            if len(self.inputs.in_files) > 1:
                # for multi-image segmentation there is one bias field image
                # per input
                for val, f in enumerate(self.inputs.in_files):
                    # image numbering is 1-based
                    outputs["bias_field"].append(
                        self._gen_fname(
                            suffix="_bias_%d" % (val + 1), **_gen_fname_opts
                        )
                    )
            else:
                # single image segmentation has unnumbered output image
                outputs["bias_field"].append(
                    self._gen_fname(suffix="_bias", **_gen_fname_opts)
                )

        if self.inputs.probability_maps:
            outputs["probability_maps"] = []
            for i in range(nclasses):
                outputs["probability_maps"].append(
                    self._gen_fname(suffix="_prob_%d" % i, **_gen_fname_opts)
                )
        return outputs


class FLIRTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True, argstr="-in %s", mandatory=True, position=0, desc="input file"
    )
    reference = File(
        exists=True, argstr="-ref %s", mandatory=True, position=1, desc="reference file"
    )
    out_file = File(
        argstr="-out %s",
        desc="registered output file",
        name_source=["in_file"],
        name_template="%s_flirt",
        position=2,
        hash_files=False,
    )
    out_matrix_file = File(
        argstr="-omat %s",
        name_source=["in_file"],
        keep_extension=True,
        name_template="%s_flirt.mat",
        desc="output affine matrix in 4x4 asciii format",
        position=3,
        hash_files=False,
    )
    out_log = File(
        name_source=["in_file"],
        keep_extension=True,
        requires=["save_log"],
        name_template="%s_flirt.log",
        desc="output log",
    )
    in_matrix_file = File(argstr="-init %s", desc="input 4x4 affine matrix")
    apply_xfm = traits.Bool(
        argstr="-applyxfm",
        desc=(
            "apply transformation supplied by in_matrix_file or uses_qform to"
            " use the affine matrix stored in the reference header"
        ),
    )
    apply_isoxfm = traits.Float(
        argstr="-applyisoxfm %f",
        xor=["apply_xfm"],
        desc="as applyxfm but forces isotropic resampling",
    )
    datatype = traits.Enum(
        "char",
        "short",
        "int",
        "float",
        "double",
        argstr="-datatype %s",
        desc="force output data type",
    )
    cost = traits.Enum(
        "mutualinfo",
        "corratio",
        "normcorr",
        "normmi",
        "leastsq",
        "labeldiff",
        "bbr",
        argstr="-cost %s",
        desc="cost function",
    )
    # XXX What is the difference between 'cost' and 'searchcost'?  Are
    # these both necessary or do they map to the same variable.
    cost_func = traits.Enum(
        "mutualinfo",
        "corratio",
        "normcorr",
        "normmi",
        "leastsq",
        "labeldiff",
        "bbr",
        argstr="-searchcost %s",
        desc="cost function",
    )
    uses_qform = traits.Bool(
        argstr="-usesqform", desc="initialize using sform or qform"
    )
    display_init = traits.Bool(argstr="-displayinit", desc="display initial matrix")
    angle_rep = traits.Enum(
        "quaternion",
        "euler",
        argstr="-anglerep %s",
        desc="representation of rotation angles",
    )
    interp = traits.Enum(
        "trilinear",
        "nearestneighbour",
        "sinc",
        "spline",
        argstr="-interp %s",
        desc="final interpolation method used in reslicing",
    )
    sinc_width = traits.Int(
        argstr="-sincwidth %d", units="voxels", desc="full-width in voxels"
    )
    sinc_window = traits.Enum(
        "rectangular",
        "hanning",
        "blackman",
        argstr="-sincwindow %s",
        desc="sinc window",
    )  # XXX better doc
    bins = traits.Int(argstr="-bins %d", desc="number of histogram bins")
    dof = traits.Int(argstr="-dof %d", desc="number of transform degrees of freedom")
    no_resample = traits.Bool(argstr="-noresample", desc="do not change input sampling")
    force_scaling = traits.Bool(
        argstr="-forcescaling", desc="force rescaling even for low-res images"
    )
    min_sampling = traits.Float(
        argstr="-minsampling %f",
        units="mm",
        desc="set minimum voxel dimension for sampling",
    )
    padding_size = traits.Int(
        argstr="-paddingsize %d",
        units="voxels",
        desc="for applyxfm: interpolates outside image " "by size",
    )
    searchr_x = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units="degrees",
        argstr="-searchrx %s",
        desc="search angles along x-axis, in degrees",
    )
    searchr_y = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units="degrees",
        argstr="-searchry %s",
        desc="search angles along y-axis, in degrees",
    )
    searchr_z = traits.List(
        traits.Int,
        minlen=2,
        maxlen=2,
        units="degrees",
        argstr="-searchrz %s",
        desc="search angles along z-axis, in degrees",
    )
    no_search = traits.Bool(
        argstr="-nosearch", desc="set all angular searches to ranges 0 to 0"
    )
    coarse_search = traits.Int(
        argstr="-coarsesearch %d", units="degrees", desc="coarse search delta angle"
    )
    fine_search = traits.Int(
        argstr="-finesearch %d", units="degrees", desc="fine search delta angle"
    )
    schedule = File(
        exists=True, argstr="-schedule %s", desc="replaces default schedule"
    )
    ref_weight = File(
        exists=True, argstr="-refweight %s", desc="File for reference weighting volume"
    )
    in_weight = File(
        exists=True, argstr="-inweight %s", desc="File for input weighting volume"
    )
    no_clamp = traits.Bool(argstr="-noclamp", desc="do not use intensity clamping")
    no_resample_blur = traits.Bool(
        argstr="-noresampblur", desc="do not use blurring on downsampling"
    )
    rigid2D = traits.Bool(argstr="-2D", desc="use 2D rigid body mode - ignores dof")
    save_log = traits.Bool(desc="save to log file")
    verbose = traits.Int(argstr="-verbose %d", desc="verbose mode, 0 is least")
    bgvalue = traits.Float(
        0,
        argstr="-setbackground %f",
        desc=("use specified background value for points " "outside FOV"),
    )

    # BBR options
    wm_seg = File(
        argstr="-wmseg %s",
        min_ver="5.0.0",
        desc="white matter segmentation volume needed by BBR cost function",
    )
    wmcoords = File(
        argstr="-wmcoords %s",
        min_ver="5.0.0",
        desc="white matter boundary coordinates for BBR cost function",
    )
    wmnorms = File(
        argstr="-wmnorms %s",
        min_ver="5.0.0",
        desc="white matter boundary normals for BBR cost function",
    )
    fieldmap = File(
        argstr="-fieldmap %s",
        min_ver="5.0.0",
        desc=(
            "fieldmap image in rads/s - must be already registered to the "
            "reference image"
        ),
    )
    fieldmapmask = File(
        argstr="-fieldmapmask %s", min_ver="5.0.0", desc="mask for fieldmap image"
    )
    pedir = traits.Int(
        argstr="-pedir %d",
        min_ver="5.0.0",
        desc="phase encode direction of EPI - 1/2/3=x/y/z & -1/-2/-3=-x/-y/-z",
    )
    echospacing = traits.Float(
        argstr="-echospacing %f",
        min_ver="5.0.0",
        desc="value of EPI echo spacing - units of seconds",
    )
    bbrtype = traits.Enum(
        "signed",
        "global_abs",
        "local_abs",
        argstr="-bbrtype %s",
        min_ver="5.0.0",
        desc=("type of bbr cost function: signed [default], global_abs, " "local_abs"),
    )
    bbrslope = traits.Float(
        argstr="-bbrslope %f", min_ver="5.0.0", desc="value of bbr slope"
    )


class FLIRTOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="path/name of registered file (if generated)")
    out_matrix_file = File(
        exists=True, desc="path/name of calculated affine transform " "(if generated)"
    )
    out_log = File(desc="path/name of output log (if generated)")


class FLIRT(FSLCommand):
    """FSL FLIRT wrapper for coregistration

    For complete details, see the `FLIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FLIRT>`_

    To print out the command line help, use:
        fsl.FLIRT().inputs_help()

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> flt = fsl.FLIRT(bins=640, cost_func='mutualinfo')
    >>> flt.inputs.in_file = 'structural.nii'
    >>> flt.inputs.reference = 'mni.nii'
    >>> flt.inputs.output_type = "NIFTI_GZ"
    >>> flt.cmdline # doctest: +ELLIPSIS
    'flirt -in structural.nii -ref mni.nii -out structural_flirt.nii.gz -omat structural_flirt.mat -bins 640 -searchcost mutualinfo'
    >>> res = flt.run() #doctest: +SKIP

    """

    _cmd = "flirt"
    input_spec = FLIRTInputSpec
    output_spec = FLIRTOutputSpec
    _log_written = False

    def aggregate_outputs(self, runtime=None, needed_outputs=None):
        outputs = super(FLIRT, self).aggregate_outputs(
            runtime=runtime, needed_outputs=needed_outputs
        )
        if self.inputs.save_log and not self._log_written:
            with open(outputs.out_log, "a") as text_file:
                text_file.write(runtime.stdout + "\n")
            self._log_written = True
        return outputs

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []
        if self.inputs.save_log and not self.inputs.verbose:
            self.inputs.verbose = 1
        if self.inputs.apply_xfm and not (
            self.inputs.in_matrix_file or self.inputs.uses_qform
        ):
            raise RuntimeError(
                "Argument apply_xfm requires in_matrix_file or "
                "uses_qform arguments to run"
            )
        skip.append("save_log")
        return super(FLIRT, self)._parse_inputs(skip=skip)


class ApplyXFMInputSpec(FLIRTInputSpec):
    apply_xfm = traits.Bool(
        True,
        argstr="-applyxfm",
        desc=(
            "apply transformation supplied by in_matrix_file or uses_qform to"
            " use the affine matrix stored in the reference header"
        ),
        usedefault=True,
    )


class ApplyXFM(FLIRT):
    """Currently just a light wrapper around FLIRT,
    with no modifications

    ApplyXFM is used to apply an existing tranform to an image


    Examples
    --------

    >>> import nipype.interfaces.fsl as fsl
    >>> from nipype.testing import example_data
    >>> applyxfm = fsl.preprocess.ApplyXFM()
    >>> applyxfm.inputs.in_file = example_data('structural.nii')
    >>> applyxfm.inputs.in_matrix_file = example_data('trans.mat')
    >>> applyxfm.inputs.out_file = 'newfile.nii'
    >>> applyxfm.inputs.reference = example_data('mni.nii')
    >>> applyxfm.inputs.apply_xfm = True
    >>> result = applyxfm.run() # doctest: +SKIP

    """

    input_spec = ApplyXFMInputSpec


class MCFLIRTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        position=0,
        argstr="-in %s",
        mandatory=True,
        desc="timeseries to motion-correct",
    )
    out_file = File(
        argstr="-out %s", genfile=True, desc="file to write", hash_files=False
    )
    cost = traits.Enum(
        "mutualinfo",
        "woods",
        "corratio",
        "normcorr",
        "normmi",
        "leastsquares",
        argstr="-cost %s",
        desc="cost function to optimize",
    )
    bins = traits.Int(argstr="-bins %d", desc="number of histogram bins")
    dof = traits.Int(argstr="-dof %d", desc="degrees of freedom for the transformation")
    ref_vol = traits.Int(argstr="-refvol %d", desc="volume to align frames to")
    scaling = traits.Float(argstr="-scaling %.2f", desc="scaling factor to use")
    smooth = traits.Float(
        argstr="-smooth %.2f", desc="smoothing factor for the cost function"
    )
    rotation = traits.Int(
        argstr="-rotation %d", desc="scaling factor for rotation tolerances"
    )
    stages = traits.Int(
        argstr="-stages %d",
        desc="stages (if 4, perform final search with sinc interpolation",
    )
    init = File(exists=True, argstr="-init %s", desc="inital transformation matrix")
    interpolation = traits.Enum(
        "spline",
        "nn",
        "sinc",
        argstr="-%s_final",
        desc="interpolation method for transformation",
    )
    use_gradient = traits.Bool(argstr="-gdt", desc="run search on gradient images")
    use_contour = traits.Bool(argstr="-edge", desc="run search on contour images")
    mean_vol = traits.Bool(argstr="-meanvol", desc="register to mean volume")
    stats_imgs = traits.Bool(
        argstr="-stats", desc="produce variance and std. dev. images"
    )
    save_mats = traits.Bool(argstr="-mats", desc="save transformation matrices")
    save_plots = traits.Bool(argstr="-plots", desc="save transformation parameters")
    save_rms = traits.Bool(
        argstr="-rmsabs -rmsrel", desc="save rms displacement parameters"
    )
    ref_file = File(
        exists=True, argstr="-reffile %s", desc="target image for motion correction"
    )


class MCFLIRTOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="motion-corrected timeseries")
    variance_img = File(exists=True, desc="variance image")
    std_img = File(exists=True, desc="standard deviation image")
    mean_img = File(exists=True, desc="mean timeseries image (if mean_vol=True)")
    par_file = File(exists=True, desc="text-file with motion parameters")
    mat_file = OutputMultiPath(File(exists=True), desc="transformation matrices")
    rms_files = OutputMultiPath(
        File(exists=True), desc="absolute and relative displacement parameters"
    )


class MCFLIRT(FSLCommand):
    """FSL MCFLIRT wrapper for within-modality motion correction

    For complete details, see the `MCFLIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/MCFLIRT>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> mcflt = fsl.MCFLIRT()
    >>> mcflt.inputs.in_file = 'functional.nii'
    >>> mcflt.inputs.cost = 'mutualinfo'
    >>> mcflt.inputs.out_file = 'moco.nii'
    >>> mcflt.cmdline
    'mcflirt -in functional.nii -cost mutualinfo -out moco.nii'
    >>> res = mcflt.run()  # doctest: +SKIP

    """

    _cmd = "mcflirt"
    input_spec = MCFLIRTInputSpec
    output_spec = MCFLIRTOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "interpolation":
            if value == "trilinear":
                return ""
            else:
                return spec.argstr % value
        return super(MCFLIRT, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()

        outputs["out_file"] = self._gen_outfilename()
        output_dir = os.path.dirname(outputs["out_file"])

        if isdefined(self.inputs.stats_imgs) and self.inputs.stats_imgs:
            if LooseVersion(Info.version()) < LooseVersion("6.0.0"):
                # FSL <6.0 outputs have .nii.gz_variance.nii.gz as extension
                outputs["variance_img"] = self._gen_fname(
                    outputs["out_file"] + "_variance.ext", cwd=output_dir
                )
                outputs["std_img"] = self._gen_fname(
                    outputs["out_file"] + "_sigma.ext", cwd=output_dir
                )
            else:
                outputs["variance_img"] = self._gen_fname(
                    outputs["out_file"], suffix="_variance", cwd=output_dir
                )
                outputs["std_img"] = self._gen_fname(
                    outputs["out_file"], suffix="_sigma", cwd=output_dir
                )

        # The mean image created if -stats option is specified ('meanvol')
        # is missing the top and bottom slices. Therefore we only expose the
        # mean image created by -meanvol option ('mean_reg') which isn't
        # corrupted.
        # Note that the same problem holds for the std and variance image.

        if isdefined(self.inputs.mean_vol) and self.inputs.mean_vol:
            if LooseVersion(Info.version()) < LooseVersion("6.0.0"):
                # FSL <6.0 outputs have .nii.gz_mean_img.nii.gz as extension
                outputs["mean_img"] = self._gen_fname(
                    outputs["out_file"] + "_mean_reg.ext", cwd=output_dir
                )
            else:
                outputs["mean_img"] = self._gen_fname(
                    outputs["out_file"], suffix="_mean_reg", cwd=output_dir
                )

        if isdefined(self.inputs.save_mats) and self.inputs.save_mats:
            _, filename = os.path.split(outputs["out_file"])
            matpathname = os.path.join(output_dir, filename + ".mat")
            _, _, _, timepoints = load(self.inputs.in_file).shape
            outputs["mat_file"] = []
            for t in range(timepoints):
                outputs["mat_file"].append(os.path.join(matpathname, "MAT_%04d" % t))
        if isdefined(self.inputs.save_plots) and self.inputs.save_plots:
            # Note - if e.g. out_file has .nii.gz, you get .nii.gz.par,
            # which is what mcflirt does!
            outputs["par_file"] = outputs["out_file"] + ".par"
        if isdefined(self.inputs.save_rms) and self.inputs.save_rms:
            outfile = outputs["out_file"]
            outputs["rms_files"] = [outfile + "_abs.rms", outfile + "_rel.rms"]
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._gen_outfilename()
        return None

    def _gen_outfilename(self):
        out_file = self.inputs.out_file
        if isdefined(out_file):
            out_file = os.path.realpath(out_file)
        if not isdefined(out_file) and isdefined(self.inputs.in_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix="_mcf")
        return os.path.abspath(out_file)


class FNIRTInputSpec(FSLCommandInputSpec):
    ref_file = File(
        exists=True, argstr="--ref=%s", mandatory=True, desc="name of reference image"
    )
    in_file = File(
        exists=True, argstr="--in=%s", mandatory=True, desc="name of input image"
    )
    affine_file = File(
        exists=True, argstr="--aff=%s", desc="name of file containing affine transform"
    )
    inwarp_file = File(
        exists=True,
        argstr="--inwarp=%s",
        desc="name of file containing initial non-linear warps",
    )
    in_intensitymap_file = traits.List(
        File(exists=True),
        argstr="--intin=%s",
        copyfile=False,
        minlen=1,
        maxlen=2,
        desc=(
            "name of file/files containing "
            "initial intensity mapping "
            "usually generated by previous "
            "fnirt run"
        ),
    )
    fieldcoeff_file = traits.Either(
        traits.Bool,
        File,
        argstr="--cout=%s",
        desc="name of output file with field coefficients or true",
    )
    warped_file = File(
        argstr="--iout=%s", desc="name of output image", genfile=True, hash_files=False
    )
    field_file = traits.Either(
        traits.Bool,
        File,
        argstr="--fout=%s",
        desc="name of output file with field or true",
        hash_files=False,
    )
    jacobian_file = traits.Either(
        traits.Bool,
        File,
        argstr="--jout=%s",
        desc=(
            "name of file for writing out the "
            "Jacobian of the field (for "
            "diagnostic or VBM purposes)"
        ),
        hash_files=False,
    )
    modulatedref_file = traits.Either(
        traits.Bool,
        File,
        argstr="--refout=%s",
        desc=(
            "name of file for writing out "
            "intensity modulated --ref (for "
            "diagnostic purposes)"
        ),
        hash_files=False,
    )
    out_intensitymap_file = traits.Either(
        traits.Bool,
        File,
        argstr="--intout=%s",
        desc=(
            "name of files for writing "
            "information pertaining to "
            "intensity mapping"
        ),
        hash_files=False,
    )
    log_file = File(
        argstr="--logout=%s", desc="Name of log-file", genfile=True, hash_files=False
    )
    config_file = traits.Either(
        traits.Enum("T1_2_MNI152_2mm", "FA_2_FMRIB58_1mm"),
        File(exists=True),
        argstr="--config=%s",
        desc="Name of config file specifying command line arguments",
    )
    refmask_file = File(
        exists=True,
        argstr="--refmask=%s",
        desc="name of file with mask in reference space",
    )
    inmask_file = File(
        exists=True,
        argstr="--inmask=%s",
        desc="name of file with mask in input image space",
    )
    skip_refmask = traits.Bool(
        argstr="--applyrefmask=0",
        xor=["apply_refmask"],
        desc="Skip specified refmask if set, default false",
    )
    skip_inmask = traits.Bool(
        argstr="--applyinmask=0",
        xor=["apply_inmask"],
        desc="skip specified inmask if set, default false",
    )
    apply_refmask = traits.List(
        traits.Enum(0, 1),
        argstr="--applyrefmask=%s",
        xor=["skip_refmask"],
        desc=("list of iterations to use reference mask on (1 to use, 0 to " "skip)"),
        sep=",",
    )
    apply_inmask = traits.List(
        traits.Enum(0, 1),
        argstr="--applyinmask=%s",
        xor=["skip_inmask"],
        desc="list of iterations to use input mask on (1 to use, 0 to skip)",
        sep=",",
    )
    skip_implicit_ref_masking = traits.Bool(
        argstr="--imprefm=0",
        desc=("skip implicit masking  based on value in --ref image. " "Default = 0"),
    )
    skip_implicit_in_masking = traits.Bool(
        argstr="--impinm=0",
        desc=("skip implicit masking  based on value in --in image. " "Default = 0"),
    )
    refmask_val = traits.Float(
        argstr="--imprefval=%f", desc="Value to mask out in --ref image. Default =0.0"
    )
    inmask_val = traits.Float(
        argstr="--impinval=%f", desc="Value to mask out in --in image. Default =0.0"
    )
    max_nonlin_iter = traits.List(
        traits.Int,
        argstr="--miter=%s",
        desc="Max # of non-linear iterations list, default [5, 5, 5, 5]",
        sep=",",
    )
    subsampling_scheme = traits.List(
        traits.Int,
        argstr="--subsamp=%s",
        desc="sub-sampling scheme, list, default [4, 2, 1, 1]",
        sep=",",
    )
    warp_resolution = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        argstr="--warpres=%d,%d,%d",
        desc=(
            "(approximate) resolution (in mm) of warp basis in x-, y- and "
            "z-direction, default 10, 10, 10"
        ),
    )
    spline_order = traits.Int(
        argstr="--splineorder=%d",
        desc="Order of spline, 2->Qadratic spline, 3->Cubic spline. Default=3",
    )
    in_fwhm = traits.List(
        traits.Int,
        argstr="--infwhm=%s",
        desc=(
            "FWHM (in mm) of gaussian smoothing kernel for input volume, "
            "default [6, 4, 2, 2]"
        ),
        sep=",",
    )
    ref_fwhm = traits.List(
        traits.Int,
        argstr="--reffwhm=%s",
        desc=(
            "FWHM (in mm) of gaussian smoothing kernel for ref volume, "
            "default [4, 2, 0, 0]"
        ),
        sep=",",
    )
    regularization_model = traits.Enum(
        "membrane_energy",
        "bending_energy",
        argstr="--regmod=%s",
        desc=(
            "Model for regularisation of warp-field [membrane_energy "
            "bending_energy], default bending_energy"
        ),
    )
    regularization_lambda = traits.List(
        traits.Float,
        argstr="--lambda=%s",
        desc=(
            "Weight of regularisation, default depending on --ssqlambda and "
            "--regmod switches. See user documetation."
        ),
        sep=",",
    )
    skip_lambda_ssq = traits.Bool(
        argstr="--ssqlambda=0",
        desc="If true, lambda is not weighted by current ssq, default false",
    )
    jacobian_range = traits.Tuple(
        traits.Float,
        traits.Float,
        argstr="--jacrange=%f,%f",
        desc="Allowed range of Jacobian determinants, default 0.01, 100.0",
    )
    derive_from_ref = traits.Bool(
        argstr="--refderiv",
        desc=("If true, ref image is used to calculate derivatives. " "Default false"),
    )
    intensity_mapping_model = traits.Enum(
        "none",
        "global_linear",
        "global_non_linear",
        "local_linear",
        "global_non_linear_with_bias",
        "local_non_linear",
        argstr="--intmod=%s",
        desc="Model for intensity-mapping",
    )
    intensity_mapping_order = traits.Int(
        argstr="--intorder=%d",
        desc="Order of poynomial for mapping intensities, default 5",
    )
    biasfield_resolution = traits.Tuple(
        traits.Int,
        traits.Int,
        traits.Int,
        argstr="--biasres=%d,%d,%d",
        desc=(
            "Resolution (in mm) of bias-field modelling local intensities, "
            "default 50, 50, 50"
        ),
    )
    bias_regularization_lambda = traits.Float(
        argstr="--biaslambda=%f",
        desc="Weight of regularisation for bias-field, default 10000",
    )
    skip_intensity_mapping = traits.Bool(
        argstr="--estint=0",
        xor=["apply_intensity_mapping"],
        desc="Skip estimate intensity-mapping default false",
    )
    apply_intensity_mapping = traits.List(
        traits.Enum(0, 1),
        argstr="--estint=%s",
        xor=["skip_intensity_mapping"],
        desc=(
            "List of subsampling levels to apply intensity mapping for "
            "(0 to skip, 1 to apply)"
        ),
        sep=",",
    )
    hessian_precision = traits.Enum(
        "double",
        "float",
        argstr="--numprec=%s",
        desc=("Precision for representing Hessian, double or float. " "Default double"),
    )


class FNIRTOutputSpec(TraitedSpec):
    fieldcoeff_file = File(exists=True, desc="file with field coefficients")
    warped_file = File(exists=True, desc="warped image")
    field_file = File(desc="file with warp field")
    jacobian_file = File(desc="file containing Jacobian of the field")
    modulatedref_file = File(desc="file containing intensity modulated --ref")
    out_intensitymap_file = traits.List(
        File,
        minlen=2,
        maxlen=2,
        desc="files containing info pertaining to intensity mapping",
    )
    log_file = File(desc="Name of log-file")


class FNIRT(FSLCommand):
    """FSL FNIRT wrapper for non-linear registration

    For complete details, see the `FNIRT Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FNIRT>`_

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> fnt = fsl.FNIRT(affine_file=example_data('trans.mat'))
    >>> res = fnt.run(ref_file=example_data('mni.nii', in_file=example_data('structural.nii')) #doctest: +SKIP

    T1 -> Mni153

    >>> from nipype.interfaces import fsl
    >>> fnirt_mprage = fsl.FNIRT()
    >>> fnirt_mprage.inputs.in_fwhm = [8, 4, 2, 2]
    >>> fnirt_mprage.inputs.subsampling_scheme = [4, 2, 1, 1]

    Specify the resolution of the warps

    >>> fnirt_mprage.inputs.warp_resolution = (6, 6, 6)
    >>> res = fnirt_mprage.run(in_file='structural.nii', ref_file='mni.nii', warped_file='warped.nii', fieldcoeff_file='fieldcoeff.nii')#doctest: +SKIP

    We can check the command line and confirm that it's what we expect.

    >>> fnirt_mprage.cmdline  #doctest: +SKIP
    'fnirt --cout=fieldcoeff.nii --in=structural.nii --infwhm=8,4,2,2 --ref=mni.nii --subsamp=4,2,1,1 --warpres=6,6,6 --iout=warped.nii'

    """

    _cmd = "fnirt"
    input_spec = FNIRTInputSpec
    output_spec = FNIRTOutputSpec

    filemap = {
        "warped_file": "warped",
        "field_file": "field",
        "jacobian_file": "field_jacobian",
        "modulatedref_file": "modulated",
        "out_intensitymap_file": "intmap",
        "log_file": "log.txt",
        "fieldcoeff_file": "fieldwarp",
    }

    def _list_outputs(self):
        outputs = self.output_spec().get()
        for key, suffix in list(self.filemap.items()):
            inval = getattr(self.inputs, key)
            change_ext = True
            if key in ["warped_file", "log_file"]:
                if suffix.endswith(".txt"):
                    change_ext = False
                if isdefined(inval):
                    outputs[key] = os.path.abspath(inval)
                else:
                    outputs[key] = self._gen_fname(
                        self.inputs.in_file, suffix="_" + suffix, change_ext=change_ext
                    )
            elif isdefined(inval):
                if isinstance(inval, bool):
                    if inval:
                        outputs[key] = self._gen_fname(
                            self.inputs.in_file,
                            suffix="_" + suffix,
                            change_ext=change_ext,
                        )
                else:
                    outputs[key] = os.path.abspath(inval)

            if key == "out_intensitymap_file" and isdefined(outputs[key]):
                basename = FNIRT.intensitymap_file_basename(outputs[key])
                outputs[key] = [outputs[key], "%s.txt" % basename]
        return outputs

    def _format_arg(self, name, spec, value):
        if name in ("in_intensitymap_file", "out_intensitymap_file"):
            if name == "out_intensitymap_file":
                value = self._list_outputs()[name]
            value = [FNIRT.intensitymap_file_basename(v) for v in value]
            assert len(set(value)) == 1, "Found different basenames for {}: {}".format(
                name, value
            )
            return spec.argstr % value[0]
        if name in list(self.filemap.keys()):
            return spec.argstr % self._list_outputs()[name]
        return super(FNIRT, self)._format_arg(name, spec, value)

    def _gen_filename(self, name):
        if name in ["warped_file", "log_file"]:
            return self._list_outputs()[name]
        return None

    def write_config(self, configfile):
        """Writes out currently set options to specified config file

        XX TODO : need to figure out how the config file is written

        Parameters
        ----------
        configfile : /path/to/configfile
        """
        try:
            fid = open(configfile, "w+")
        except IOError:
            print("unable to create config_file %s" % (configfile))

        for item in list(self.inputs.get().items()):
            fid.write("%s\n" % (item))
        fid.close()

    @classmethod
    def intensitymap_file_basename(cls, f):
        """Removes valid intensitymap extensions from `f`, returning a basename
        that can refer to both intensitymap files.
        """
        for ext in list(Info.ftypes.values()) + [".txt"]:
            if f.endswith(ext):
                return f[: -len(ext)]
        # TODO consider warning for this case
        return f


class ApplyWarpInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr="--in=%s",
        mandatory=True,
        position=0,
        desc="image to be warped",
    )
    out_file = File(
        argstr="--out=%s",
        genfile=True,
        position=2,
        desc="output filename",
        hash_files=False,
    )
    ref_file = File(
        exists=True,
        argstr="--ref=%s",
        mandatory=True,
        position=1,
        desc="reference image",
    )
    field_file = File(
        exists=True, argstr="--warp=%s", desc="file containing warp field"
    )
    abswarp = traits.Bool(
        argstr="--abs", xor=["relwarp"], desc="treat warp field as absolute: x' = w(x)"
    )
    relwarp = traits.Bool(
        argstr="--rel",
        xor=["abswarp"],
        position=-1,
        desc="treat warp field as relative: x' = x + w(x)",
    )
    datatype = traits.Enum(
        "char",
        "short",
        "int",
        "float",
        "double",
        argstr="--datatype=%s",
        desc="Force output data type [char short int float double].",
    )
    supersample = traits.Bool(
        argstr="--super", desc="intermediary supersampling of output, default is off"
    )
    superlevel = traits.Either(
        traits.Enum("a"),
        traits.Int,
        argstr="--superlevel=%s",
        desc=(
            "level of intermediary supersampling, a for 'automatic' or "
            "integer level. Default = 2"
        ),
    )
    premat = File(
        exists=True,
        argstr="--premat=%s",
        desc="filename for pre-transform (affine matrix)",
    )
    postmat = File(
        exists=True,
        argstr="--postmat=%s",
        desc="filename for post-transform (affine matrix)",
    )
    mask_file = File(
        exists=True,
        argstr="--mask=%s",
        desc="filename for mask image (in reference space)",
    )
    interp = traits.Enum(
        "nn",
        "trilinear",
        "sinc",
        "spline",
        argstr="--interp=%s",
        position=-2,
        desc="interpolation method",
    )


class ApplyWarpOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="Warped output file")


class ApplyWarp(FSLCommand):
    """FSL's applywarp wrapper to apply the results of a FNIRT registration

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> aw = fsl.ApplyWarp()
    >>> aw.inputs.in_file = example_data('structural.nii')
    >>> aw.inputs.ref_file = example_data('mni.nii')
    >>> aw.inputs.field_file = 'my_coefficients_filed.nii' #doctest: +SKIP
    >>> res = aw.run() #doctest: +SKIP


    """

    _cmd = "applywarp"
    input_spec = ApplyWarpInputSpec
    output_spec = ApplyWarpOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "superlevel":
            return spec.argstr % str(value)
        return super(ApplyWarp, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        if not isdefined(self.inputs.out_file):
            outputs["out_file"] = self._gen_fname(self.inputs.in_file, suffix="_warp")
        else:
            outputs["out_file"] = os.path.abspath(self.inputs.out_file)
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()[name]
        return None


class SliceTimerInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr="--in=%s",
        mandatory=True,
        position=0,
        desc="filename of input timeseries",
    )
    out_file = File(
        argstr="--out=%s",
        genfile=True,
        desc="filename of output timeseries",
        hash_files=False,
    )
    index_dir = traits.Bool(argstr="--down", desc="slice indexing from top to bottom")
    time_repetition = traits.Float(
        argstr="--repeat=%f", desc="Specify TR of data - default is 3s"
    )
    slice_direction = traits.Enum(
        1,
        2,
        3,
        argstr="--direction=%d",
        desc="direction of slice acquisition (x=1, y=2, z=3) - default is z",
    )
    interleaved = traits.Bool(argstr="--odd", desc="use interleaved acquisition")
    custom_timings = File(
        exists=True,
        argstr="--tcustom=%s",
        desc=(
            "slice timings, in fractions of TR, range 0:1 (default is 0.5 = "
            "no shift)"
        ),
    )
    global_shift = traits.Float(
        argstr="--tglobal",
        desc="shift in fraction of TR, range 0:1 (default is 0.5 = no shift)",
    )
    custom_order = File(
        exists=True,
        argstr="--ocustom=%s",
        desc=(
            "filename of single-column custom interleave order file (first "
            "slice is referred to as 1 not 0)"
        ),
    )


class SliceTimerOutputSpec(TraitedSpec):
    slice_time_corrected_file = File(exists=True, desc="slice time corrected file")


class SliceTimer(FSLCommand):
    """FSL slicetimer wrapper to perform slice timing correction

    Examples
    --------
    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> st = fsl.SliceTimer()
    >>> st.inputs.in_file = example_data('functional.nii')
    >>> st.inputs.interleaved = True
    >>> result = st.run() #doctest: +SKIP

    """

    _cmd = "slicetimer"
    input_spec = SliceTimerInputSpec
    output_spec = SliceTimerOutputSpec

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix="_st")
        outputs["slice_time_corrected_file"] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()["slice_time_corrected_file"]
        return None


class SUSANInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        argstr="%s",
        mandatory=True,
        position=1,
        desc="filename of input timeseries",
    )
    brightness_threshold = traits.Float(
        argstr="%.10f",
        position=2,
        mandatory=True,
        desc=(
            "brightness threshold and should be greater than noise level "
            "and less than contrast of edges to be preserved."
        ),
    )
    fwhm = traits.Float(
        argstr="%.10f",
        position=3,
        mandatory=True,
        desc="fwhm of smoothing, in mm, gets converted using sqrt(8*log(2))",
    )
    dimension = traits.Enum(
        3,
        2,
        argstr="%d",
        position=4,
        usedefault=True,
        desc="within-plane (2) or fully 3D (3)",
    )
    use_median = traits.Enum(
        1,
        0,
        argstr="%d",
        position=5,
        usedefault=True,
        desc=(
            "whether to use a local median filter in the cases where "
            "single-point noise is detected"
        ),
    )
    usans = traits.List(
        traits.Tuple(File(exists=True), traits.Float),
        maxlen=2,
        argstr="",
        position=6,
        usedefault=True,
        desc="determines whether the smoothing area (USAN) is to be "
        "found from secondary images (0, 1 or 2). A negative "
        "value for any brightness threshold will auto-set the "
        "threshold at 10% of the robust range",
    )
    out_file = File(
        argstr="%s",
        position=-1,
        genfile=True,
        desc="output file name",
        hash_files=False,
    )


class SUSANOutputSpec(TraitedSpec):
    smoothed_file = File(exists=True, desc="smoothed output file")


class SUSAN(FSLCommand):
    """FSL SUSAN wrapper to perform smoothing

    For complete details, see the `SUSAN Documentation.
    <https://fsl.fmrib.ox.ac.uk/fsl/fslwiki/SUSAN>`_

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> from nipype.testing import example_data
    >>> anatfile  # doctest: +SKIP
    anatomical.nii  # doctest: +SKIP
    >>> sus = fsl.SUSAN()
    >>> sus.inputs.in_file = example_data('structural.nii')
    >>> sus.inputs.brightness_threshold = 2000.0
    >>> sus.inputs.fwhm = 8.0
    >>> result = sus.run()  # doctest: +SKIP
    """

    _cmd = "susan"
    input_spec = SUSANInputSpec
    output_spec = SUSANOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "fwhm":
            return spec.argstr % (float(value) / np.sqrt(8 * np.log(2)))
        if name == "usans":
            if not value:
                return "0"
            arglist = [str(len(value))]
            for filename, thresh in value:
                arglist.extend([filename, "%.10f" % thresh])
            return " ".join(arglist)
        return super(SUSAN, self)._format_arg(name, spec, value)

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.out_file
        if not isdefined(out_file):
            out_file = self._gen_fname(self.inputs.in_file, suffix="_smooth")
        outputs["smoothed_file"] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            return self._list_outputs()["smoothed_file"]
        return None


class FUGUEInputSpec(FSLCommandInputSpec):
    in_file = File(exists=True, argstr="--in=%s", desc="filename of input volume")
    shift_in_file = File(
        exists=True,
        argstr="--loadshift=%s",
        desc="filename for reading pixel shift volume",
    )
    phasemap_in_file = File(
        exists=True, argstr="--phasemap=%s", desc="filename for input phase image"
    )
    fmap_in_file = File(
        exists=True,
        argstr="--loadfmap=%s",
        desc="filename for loading fieldmap (rad/s)",
    )
    unwarped_file = File(
        argstr="--unwarp=%s",
        desc="apply unwarping and save as filename",
        xor=["warped_file"],
        requires=["in_file"],
    )
    warped_file = File(
        argstr="--warp=%s",
        desc="apply forward warping and save as filename",
        xor=["unwarped_file"],
        requires=["in_file"],
    )
    forward_warping = traits.Bool(
        False, usedefault=True, desc="apply forward warping instead of unwarping"
    )
    dwell_to_asym_ratio = traits.Float(
        argstr="--dwelltoasym=%.10f", desc="set the dwell to asym time ratio"
    )
    dwell_time = traits.Float(
        argstr="--dwell=%.10f",
        desc=(
            "set the EPI dwell time per phase-encode line - same as echo "
            "spacing - (sec)"
        ),
    )
    asym_se_time = traits.Float(
        argstr="--asym=%.10f", desc="set the fieldmap asymmetric spin echo time (sec)"
    )
    median_2dfilter = traits.Bool(argstr="--median", desc="apply 2D median filtering")
    despike_2dfilter = traits.Bool(
        argstr="--despike", desc="apply a 2D de-spiking filter"
    )
    no_gap_fill = traits.Bool(
        argstr="--nofill", desc="do not apply gap-filling measure to the fieldmap"
    )
    no_extend = traits.Bool(
        argstr="--noextend",
        desc="do not apply rigid-body extrapolation to the fieldmap",
    )
    smooth2d = traits.Float(
        argstr="--smooth2=%.2f", desc="apply 2D Gaussian smoothing of sigma N (in mm)"
    )
    smooth3d = traits.Float(
        argstr="--smooth3=%.2f", desc="apply 3D Gaussian smoothing of sigma N (in mm)"
    )
    poly_order = traits.Int(
        argstr="--poly=%d", desc="apply polynomial fitting of order N"
    )
    fourier_order = traits.Int(
        argstr="--fourier=%d", desc="apply Fourier (sinusoidal) fitting of order N"
    )
    pava = traits.Bool(argstr="--pava", desc="apply monotonic enforcement via PAVA")
    despike_threshold = traits.Float(
        argstr="--despikethreshold=%s",
        desc="specify the threshold for de-spiking (default=3.0)",
    )
    unwarp_direction = traits.Enum(
        "x",
        "y",
        "z",
        "x-",
        "y-",
        "z-",
        argstr="--unwarpdir=%s",
        desc="specifies direction of warping (default y)",
    )
    phase_conjugate = traits.Bool(
        argstr="--phaseconj", desc="apply phase conjugate method of unwarping"
    )
    icorr = traits.Bool(
        argstr="--icorr",
        requires=["shift_in_file"],
        desc=("apply intensity correction to unwarping (pixel shift method " "only)"),
    )
    icorr_only = traits.Bool(
        argstr="--icorronly",
        requires=["unwarped_file"],
        desc="apply intensity correction only",
    )
    mask_file = File(
        exists=True, argstr="--mask=%s", desc="filename for loading valid mask"
    )
    nokspace = traits.Bool(
        False, argstr="--nokspace", desc="do not use k-space forward warping"
    )

    # Special outputs: shift (voxel shift map, vsm)
    save_shift = traits.Bool(
        False, xor=["save_unmasked_shift"], desc="write pixel shift volume"
    )
    shift_out_file = File(
        argstr="--saveshift=%s", desc="filename for saving pixel shift volume"
    )
    save_unmasked_shift = traits.Bool(
        argstr="--unmaskshift",
        xor=["save_shift"],
        desc="saves the unmasked shiftmap when using --saveshift",
    )

    # Special outputs: fieldmap (fmap)
    save_fmap = traits.Bool(
        False, xor=["save_unmasked_fmap"], desc="write field map volume"
    )
    fmap_out_file = File(
        argstr="--savefmap=%s", desc="filename for saving fieldmap (rad/s)"
    )
    save_unmasked_fmap = traits.Bool(
        False,
        argstr="--unmaskfmap",
        xor=["save_fmap"],
        desc="saves the unmasked fieldmap when using --savefmap",
    )


class FUGUEOutputSpec(TraitedSpec):
    unwarped_file = File(desc="unwarped file")
    warped_file = File(desc="forward warped file")
    shift_out_file = File(desc="voxel shift map file")
    fmap_out_file = File(desc="fieldmap file")


class FUGUE(FSLCommand):
    """FSL FUGUE set of tools for EPI distortion correction

    `FUGUE <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE>`_ is, most generally,
    a set of tools for EPI distortion correction.

    Distortions may be corrected for
        1. improving registration with non-distorted images (e.g. structurals),
           or
        2. dealing with motion-dependent changes.

    FUGUE is designed to deal only with the first case -
    improving registration.


    Examples
    --------


    Unwarping an input image (shift map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.in_file = 'epi.nii'
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.shift_in_file = 'vsm.nii'  # Previously computed with fugue as well
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --in=epi.nii --mask=epi_mask.nii --loadshift=vsm.nii --unwarpdir=y --unwarp=epi_unwarped.nii.gz'
    >>> fugue.run() #doctest: +SKIP


    Warping an input image (shift map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.in_file = 'epi.nii'
    >>> fugue.inputs.forward_warping = True
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.shift_in_file = 'vsm.nii'  # Previously computed with fugue as well
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --in=epi.nii --mask=epi_mask.nii --loadshift=vsm.nii --unwarpdir=y --warp=epi_warped.nii.gz'
    >>> fugue.run() #doctest: +SKIP


    Computing the vsm (unwrapped phase map is known):

    >>> from nipype.interfaces.fsl.preprocess import FUGUE
    >>> fugue = FUGUE()
    >>> fugue.inputs.phasemap_in_file = 'epi_phasediff.nii'
    >>> fugue.inputs.mask_file = 'epi_mask.nii'
    >>> fugue.inputs.dwell_to_asym_ratio = (0.77e-3 * 3) / 2.46e-3
    >>> fugue.inputs.unwarp_direction = 'y'
    >>> fugue.inputs.save_shift = True
    >>> fugue.inputs.output_type = "NIFTI_GZ"
    >>> fugue.cmdline # doctest: +ELLIPSIS
    'fugue --dwelltoasym=0.9390243902 --mask=epi_mask.nii --phasemap=epi_phasediff.nii --saveshift=epi_phasediff_vsm.nii.gz --unwarpdir=y'
    >>> fugue.run() #doctest: +SKIP


    """

    _cmd = "fugue"
    input_spec = FUGUEInputSpec
    output_spec = FUGUEOutputSpec

    def _parse_inputs(self, skip=None):
        if skip is None:
            skip = []

        input_phase = isdefined(self.inputs.phasemap_in_file)
        input_vsm = isdefined(self.inputs.shift_in_file)
        input_fmap = isdefined(self.inputs.fmap_in_file)

        if not input_phase and not input_vsm and not input_fmap:
            raise RuntimeError(
                (
                    "Either phasemap_in_file, shift_in_file or fmap_in_file must "
                    "be set."
                )
            )

        if not isdefined(self.inputs.in_file):
            skip += ["unwarped_file", "warped_file"]
        else:
            if self.inputs.forward_warping:
                skip += ["unwarped_file"]
                trait_spec = self.inputs.trait("warped_file")
                trait_spec.name_template = "%s_warped"
                trait_spec.name_source = "in_file"
                trait_spec.output_name = "warped_file"
            else:
                skip += ["warped_file"]
                trait_spec = self.inputs.trait("unwarped_file")
                trait_spec.name_template = "%s_unwarped"
                trait_spec.name_source = "in_file"
                trait_spec.output_name = "unwarped_file"

        # Handle shift output
        if not isdefined(self.inputs.shift_out_file):
            vsm_save_masked = (
                isdefined(self.inputs.save_shift) and self.inputs.save_shift
            )
            vsm_save_unmasked = (
                isdefined(self.inputs.save_unmasked_shift)
                and self.inputs.save_unmasked_shift
            )

            if vsm_save_masked or vsm_save_unmasked:
                trait_spec = self.inputs.trait("shift_out_file")
                trait_spec.output_name = "shift_out_file"

                if input_fmap:
                    trait_spec.name_source = "fmap_in_file"
                elif input_phase:
                    trait_spec.name_source = "phasemap_in_file"
                elif input_vsm:
                    trait_spec.name_source = "shift_in_file"
                else:
                    raise RuntimeError(
                        (
                            "Either phasemap_in_file, shift_in_file or "
                            "fmap_in_file must be set."
                        )
                    )

                if vsm_save_unmasked:
                    trait_spec.name_template = "%s_vsm_unmasked"
                else:
                    trait_spec.name_template = "%s_vsm"
            else:
                skip += ["save_shift", "save_unmasked_shift", "shift_out_file"]

        # Handle fieldmap output
        if not isdefined(self.inputs.fmap_out_file):
            fmap_save_masked = (
                isdefined(self.inputs.save_fmap) and self.inputs.save_fmap
            )
            fmap_save_unmasked = (
                isdefined(self.inputs.save_unmasked_fmap)
                and self.inputs.save_unmasked_fmap
            )

            if fmap_save_masked or fmap_save_unmasked:
                trait_spec = self.inputs.trait("fmap_out_file")
                trait_spec.output_name = "fmap_out_file"

                if input_vsm:
                    trait_spec.name_source = "shift_in_file"
                elif input_phase:
                    trait_spec.name_source = "phasemap_in_file"
                elif input_fmap:
                    trait_spec.name_source = "fmap_in_file"
                else:
                    raise RuntimeError(
                        (
                            "Either phasemap_in_file, shift_in_file or "
                            "fmap_in_file must be set."
                        )
                    )

                if fmap_save_unmasked:
                    trait_spec.name_template = "%s_fieldmap_unmasked"
                else:
                    trait_spec.name_template = "%s_fieldmap"
            else:
                skip += ["save_fmap", "save_unmasked_fmap", "fmap_out_file"]

        return super(FUGUE, self)._parse_inputs(skip=skip)


class PRELUDEInputSpec(FSLCommandInputSpec):
    complex_phase_file = File(
        exists=True,
        argstr="--complex=%s",
        mandatory=True,
        xor=["magnitude_file", "phase_file"],
        desc="complex phase input volume",
    )
    magnitude_file = File(
        exists=True,
        argstr="--abs=%s",
        mandatory=True,
        xor=["complex_phase_file"],
        desc="file containing magnitude image",
    )
    phase_file = File(
        exists=True,
        argstr="--phase=%s",
        mandatory=True,
        xor=["complex_phase_file"],
        desc="raw phase file",
    )
    unwrapped_phase_file = File(
        genfile=True,
        argstr="--unwrap=%s",
        desc="file containing unwrapepd phase",
        hash_files=False,
    )
    num_partitions = traits.Int(
        argstr="--numphasesplit=%d", desc="number of phase partitions to use"
    )
    labelprocess2d = traits.Bool(
        argstr="--labelslices", desc="does label processing in 2D (slice at a time)"
    )
    process2d = traits.Bool(
        argstr="--slices",
        xor=["labelprocess2d"],
        desc="does all processing in 2D (slice at a time)",
    )
    process3d = traits.Bool(
        argstr="--force3D",
        xor=["labelprocess2d", "process2d"],
        desc="forces all processing to be full 3D",
    )
    threshold = traits.Float(
        argstr="--thresh=%.10f", desc="intensity threshold for masking"
    )
    mask_file = File(
        exists=True, argstr="--mask=%s", desc="filename of mask input volume"
    )
    start = traits.Int(
        argstr="--start=%d", desc="first image number to process (default 0)"
    )
    end = traits.Int(
        argstr="--end=%d", desc="final image number to process (default Inf)"
    )
    savemask_file = File(
        argstr="--savemask=%s", desc="saving the mask volume", hash_files=False
    )
    rawphase_file = File(
        argstr="--rawphase=%s", desc="saving the raw phase output", hash_files=False
    )
    label_file = File(
        argstr="--labels=%s", desc="saving the area labels output", hash_files=False
    )
    removeramps = traits.Bool(
        argstr="--removeramps", desc="remove phase ramps during unwrapping"
    )


class PRELUDEOutputSpec(TraitedSpec):
    unwrapped_phase_file = File(exists=True, desc="unwrapped phase file")


class PRELUDE(FSLCommand):
    """FSL prelude wrapper for phase unwrapping

    Examples
    --------

    Please insert examples for use of this command

    """

    input_spec = PRELUDEInputSpec
    output_spec = PRELUDEOutputSpec
    _cmd = "prelude"

    def __init__(self, **kwargs):
        super(PRELUDE, self).__init__(**kwargs)
        warn("This has not been fully tested. Please report any failures.")

    def _list_outputs(self):
        outputs = self._outputs().get()
        out_file = self.inputs.unwrapped_phase_file
        if not isdefined(out_file):
            if isdefined(self.inputs.phase_file):
                out_file = self._gen_fname(self.inputs.phase_file, suffix="_unwrapped")
            elif isdefined(self.inputs.complex_phase_file):
                out_file = self._gen_fname(
                    self.inputs.complex_phase_file, suffix="_phase_unwrapped"
                )
        outputs["unwrapped_phase_file"] = os.path.abspath(out_file)
        return outputs

    def _gen_filename(self, name):
        if name == "unwrapped_phase_file":
            return self._list_outputs()["unwrapped_phase_file"]
        return None


class FIRSTInputSpec(FSLCommandInputSpec):
    in_file = File(
        exists=True,
        mandatory=True,
        position=-2,
        copyfile=False,
        argstr="-i %s",
        desc="input data file",
    )
    out_file = File(
        "segmented",
        usedefault=True,
        mandatory=True,
        position=-1,
        argstr="-o %s",
        desc="output data file",
        hash_files=False,
    )
    verbose = traits.Bool(argstr="-v", position=1, desc="Use verbose logging.")
    brain_extracted = traits.Bool(
        argstr="-b",
        position=2,
        desc="Input structural image is already brain-extracted",
    )
    no_cleanup = traits.Bool(
        argstr="-d",
        position=3,
        desc="Input structural image is already brain-extracted",
    )
    method = traits.Enum(
        "auto",
        "fast",
        "none",
        xor=["method_as_numerical_threshold"],
        argstr="-m %s",
        position=4,
        usedefault=True,
        desc=(
            "Method must be one of auto, fast, none, or it can be entered "
            "using the 'method_as_numerical_threshold' input"
        ),
    )
    method_as_numerical_threshold = traits.Float(
        argstr="-m %.4f",
        position=4,
        desc=(
            "Specify a numerical threshold value or use the 'method' input "
            "to choose auto, fast, or none"
        ),
    )
    list_of_specific_structures = traits.List(
        traits.Str,
        argstr="-s %s",
        sep=",",
        position=5,
        minlen=1,
        desc="Runs only on the specified structures (e.g. L_Hipp, R_Hipp"
        "L_Accu, R_Accu, L_Amyg, R_Amyg"
        "L_Caud, R_Caud, L_Pall, R_Pall"
        "L_Puta, R_Puta, L_Thal, R_Thal, BrStem",
    )
    affine_file = File(
        exists=True,
        position=6,
        argstr="-a %s",
        desc=(
            "Affine matrix to use (e.g. img2std.mat) (does not " "re-run registration)"
        ),
    )


class FIRSTOutputSpec(TraitedSpec):
    vtk_surfaces = OutputMultiPath(
        File(exists=True), desc="VTK format meshes for each subcortical region"
    )
    bvars = OutputMultiPath(File(exists=True), desc="bvars for each subcortical region")
    original_segmentations = File(
        exists=True,
        desc=(
            "3D image file containing the segmented regions "
            "as integer values. Uses CMA labelling"
        ),
    )
    segmentation_file = File(
        exists=True,
        desc=("4D image file containing a single volume per " "segmented region"),
    )


class FIRST(FSLCommand):
    """FSL run_first_all wrapper for segmentation of subcortical volumes

    http://www.fmrib.ox.ac.uk/fsl/first/index.html

    Examples
    --------

    >>> from nipype.interfaces import fsl
    >>> first = fsl.FIRST()
    >>> first.inputs.in_file = 'structural.nii'
    >>> first.inputs.out_file = 'segmented.nii'
    >>> res = first.run() #doctest: +SKIP

    """

    _cmd = "run_first_all"
    input_spec = FIRSTInputSpec
    output_spec = FIRSTOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()

        if isdefined(self.inputs.list_of_specific_structures):
            structures = self.inputs.list_of_specific_structures
        else:
            structures = [
                "L_Hipp",
                "R_Hipp",
                "L_Accu",
                "R_Accu",
                "L_Amyg",
                "R_Amyg",
                "L_Caud",
                "R_Caud",
                "L_Pall",
                "R_Pall",
                "L_Puta",
                "R_Puta",
                "L_Thal",
                "R_Thal",
                "BrStem",
            ]
        outputs["original_segmentations"] = self._gen_fname("original_segmentations")
        outputs["segmentation_file"] = self._gen_fname("segmentation_file")
        outputs["vtk_surfaces"] = self._gen_mesh_names("vtk_surfaces", structures)
        outputs["bvars"] = self._gen_mesh_names("bvars", structures)
        return outputs

    def _gen_fname(self, basename):
        path, outname, ext = split_filename(self.inputs.out_file)

        method = "none"
        if isdefined(self.inputs.method) and self.inputs.method != "none":
            method = "fast"
            if self.inputs.list_of_specific_structures and self.inputs.method == "auto":
                method = "none"

        if isdefined(self.inputs.method_as_numerical_threshold):
            thres = "%.4f" % self.inputs.method_as_numerical_threshold
            method = thres.replace(".", "")

        if basename == "original_segmentations":
            return op.abspath("%s_all_%s_origsegs.nii.gz" % (outname, method))
        if basename == "segmentation_file":
            return op.abspath("%s_all_%s_firstseg.nii.gz" % (outname, method))

        return None

    def _gen_mesh_names(self, name, structures):
        path, prefix, ext = split_filename(self.inputs.out_file)
        if name == "vtk_surfaces":
            vtks = list()
            for struct in structures:
                vtk = prefix + "-" + struct + "_first.vtk"
                vtks.append(op.abspath(vtk))
            return vtks
        if name == "bvars":
            bvars = list()
            for struct in structures:
                bvar = prefix + "-" + struct + "_first.bvars"
                bvars.append(op.abspath(bvar))
            return bvars
        return None



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..minc import XfmAvg


def test_XfmAvg_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        avg_linear=dict(
            argstr="-avg_linear",
        ),
        avg_nonlinear=dict(
            argstr="-avg_nonlinear",
        ),
        clobber=dict(
            argstr="-clobber",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        ignore_linear=dict(
            argstr="-ignore_linear",
        ),
        ignore_nonlinear=dict(
            argstr="-ignore_nonline",
        ),
        input_files=dict(
            argstr="%s",
            mandatory=True,
            position=-2,
            sep=" ",
        ),
        input_grid_files=dict(),
        output_file=dict(
            argstr="%s",
            extensions=None,
            genfile=True,
            position=-1,
        ),
        verbose=dict(
            argstr="-verbose",
        ),
    )
    inputs = XfmAvg.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_XfmAvg_outputs():
    output_map = dict(
        output_file=dict(
            extensions=None,
        ),
        output_grid=dict(
            extensions=None,
        ),
    )
    outputs = XfmAvg.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# -*- coding: utf8 -*-
"""Autogenerated file - DO NOT EDIT
If you spot a bug, please report it on the mailing list and/or change the generator."""

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    SEMLikeCommandLine,
    TraitedSpec,
    File,
    Directory,
    traits,
    isdefined,
    InputMultiPath,
    OutputMultiPath,
)
import os


class IntensityDifferenceMetricInputSpec(CommandLineInputSpec):
    sensitivityThreshold = traits.Float(
        desc="This parameter should be between 0 and 1, and defines how sensitive the metric should be to the intensity changes.",
        argstr="--sensitivityThreshold %f",
    )
    changingBandSize = traits.Int(
        desc="How far (in mm) from the boundary of the segmentation should the intensity changes be considered.",
        argstr="--changingBandSize %d",
    )
    baselineVolume = File(
        position=-4, desc="Baseline volume to be compared to", exists=True, argstr="%s"
    )
    baselineSegmentationVolume = File(
        position=-3,
        desc="Label volume that contains segmentation of the structure of interest in the baseline volume.",
        exists=True,
        argstr="%s",
    )
    followupVolume = File(
        position=-2,
        desc="Followup volume to be compare to the baseline",
        exists=True,
        argstr="%s",
    )
    outputVolume = traits.Either(
        traits.Bool,
        File(),
        position=-1,
        hash_files=False,
        desc="Output volume to keep the results of change quantification.",
        argstr="%s",
    )
    reportFileName = traits.Either(
        traits.Bool,
        File(),
        hash_files=False,
        desc="Report file name",
        argstr="--reportFileName %s",
    )


class IntensityDifferenceMetricOutputSpec(TraitedSpec):
    outputVolume = File(
        position=-1,
        desc="Output volume to keep the results of change quantification.",
        exists=True,
    )
    reportFileName = File(desc="Report file name", exists=True)


class IntensityDifferenceMetric(SEMLikeCommandLine):
    """title:
      Intensity Difference Change Detection (FAST)


    category:
      Quantification.ChangeQuantification


    description:
      Quantifies the changes between two spatially aligned images based on the pixel-wise difference of image intensities.


    version: 0.1

    contributor: Andrey Fedorov

    acknowledgements:

    """

    input_spec = IntensityDifferenceMetricInputSpec
    output_spec = IntensityDifferenceMetricOutputSpec
    _cmd = "IntensityDifferenceMetric "
    _outputs_filenames = {
        "outputVolume": "outputVolume.nii",
        "reportFileName": "reportFileName",
    }



# -*- coding: utf-8 -*-

import os
import glob

from ...utils.filemanip import split_filename
from ..base import (
    CommandLineInputSpec,
    CommandLine,
    traits,
    TraitedSpec,
    File,
    StdOutCommandLine,
    OutputMultiPath,
    StdOutCommandLineInputSpec,
    isdefined,
)


class Image2VoxelInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="-4dimage %s",
        mandatory=True,
        position=1,
        desc="4d image file",
    )
    # TODO convert list of files on the fly
    #    imagelist = File(exists=True, argstr='-imagelist %s',
    #                    mandatory=True, position=1,
    #                    desc='Name of a file containing a list of 3D images')
    #
    #    imageprefix = traits.Str(argstr='-imageprefix %s', position=3,
    #                    desc='Path to prepend onto filenames in the imagelist.')

    out_type = traits.Enum(
        "float",
        "char",
        "short",
        "int",
        "long",
        "double",
        argstr="-outputdatatype %s",
        position=2,
        desc='"i.e. Bfloat". Can be "char", "short", "int", "long", "float" or "double"',
        usedefault=True,
    )


class Image2VoxelOutputSpec(TraitedSpec):
    voxel_order = File(exists=True, desc="path/name of 4D volume in voxel order")


class Image2Voxel(StdOutCommandLine):
    """
    Converts Analyze / NIFTI / MHA files to voxel order.

    Converts scanner-order data in a supported image format to voxel-order data.
    Either takes a 4D file (all measurements in single image)
    or a list of 3D images.

    Examples
    --------

    >>> import nipype.interfaces.camino as cmon
    >>> img2vox = cmon.Image2Voxel()
    >>> img2vox.inputs.in_file = '4d_dwi.nii'
    >>> img2vox.run()                  # doctest: +SKIP
    """

    _cmd = "image2voxel"
    input_spec = Image2VoxelInputSpec
    output_spec = Image2VoxelOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["voxel_order"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + ".B" + self.inputs.out_type


class FSL2SchemeInputSpec(StdOutCommandLineInputSpec):
    bvec_file = File(
        exists=True,
        argstr="-bvecfile %s",
        mandatory=True,
        position=1,
        desc="b vector file",
    )

    bval_file = File(
        exists=True,
        argstr="-bvalfile %s",
        mandatory=True,
        position=2,
        desc="b value file",
    )

    numscans = traits.Int(
        argstr="-numscans %d",
        units="NA",
        desc="Output all measurements numerous (n) times, used when combining multiple scans from the same imaging session.",
    )

    interleave = traits.Bool(
        argstr="-interleave",
        desc="Interleave repeated scans. Only used with -numscans.",
    )

    bscale = traits.Float(
        argstr="-bscale %d",
        units="NA",
        desc="Scaling factor to convert the b-values into different units. Default is 10^6.",
    )

    diffusiontime = traits.Float(
        argstr="-diffusiontime %f", units="NA", desc="Diffusion time"
    )

    flipx = traits.Bool(
        argstr="-flipx", desc="Negate the x component of all the vectors."
    )
    flipy = traits.Bool(
        argstr="-flipy", desc="Negate the y component of all the vectors."
    )
    flipz = traits.Bool(
        argstr="-flipz", desc="Negate the z component of all the vectors."
    )
    usegradmod = traits.Bool(
        argstr="-usegradmod",
        desc="Use the gradient magnitude to scale b. This option has no effect if your gradient directions have unit magnitude.",
    )


class FSL2SchemeOutputSpec(TraitedSpec):
    scheme = File(exists=True, desc="Scheme file")


class FSL2Scheme(StdOutCommandLine):
    """
    Converts b-vectors and b-values from FSL format to a Camino scheme file.

    Examples
    --------

    >>> import nipype.interfaces.camino as cmon
    >>> makescheme = cmon.FSL2Scheme()
    >>> makescheme.inputs.bvec_file = 'bvecs'
    >>> makescheme.inputs.bvec_file = 'bvals'
    >>> makescheme.run()                  # doctest: +SKIP

    """

    _cmd = "fsl2scheme"
    input_spec = FSL2SchemeInputSpec
    output_spec = FSL2SchemeOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["scheme"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.bvec_file)
        return name + ".scheme"


class VtkStreamlinesInputSpec(StdOutCommandLineInputSpec):
    inputmodel = traits.Enum(
        "raw",
        "voxels",
        argstr="-inputmodel %s",
        desc="input model type (raw or voxels)",
        usedefault=True,
    )

    in_file = File(
        exists=True, argstr=" < %s", mandatory=True, position=-2, desc="data file"
    )

    voxeldims = traits.List(
        traits.Int,
        desc="voxel dimensions in mm",
        argstr="-voxeldims %s",
        minlen=3,
        maxlen=3,
        position=4,
        units="mm",
    )

    seed_file = File(
        exists=False,
        argstr="-seedfile %s",
        position=1,
        desc="image containing seed points",
    )

    target_file = File(
        exists=False,
        argstr="-targetfile %s",
        position=2,
        desc="image containing integer-valued target regions",
    )

    scalar_file = File(
        exists=False,
        argstr="-scalarfile %s",
        position=3,
        desc="image that is in the same physical space as the tracts",
    )

    colourorient = traits.Bool(
        argstr="-colourorient",
        desc="Each point on the streamline is coloured by the local orientation.",
    )
    interpolatescalars = traits.Bool(
        argstr="-interpolatescalars",
        desc="the scalar value at each point on the streamline is calculated by trilinear interpolation",
    )
    interpolate = traits.Bool(
        argstr="-interpolate",
        desc="the scalar value at each point on the streamline is calculated by trilinear interpolation",
    )


class VtkStreamlinesOutputSpec(TraitedSpec):
    vtk = File(exists=True, desc="Streamlines in VTK format")


class VtkStreamlines(StdOutCommandLine):
    """
    Use vtkstreamlines to convert raw or voxel format streamlines to VTK polydata

    Examples
    --------

    >>> import nipype.interfaces.camino as cmon
    >>> vtk = cmon.VtkStreamlines()
    >>> vtk.inputs.in_file = 'tract_data.Bfloat'
    >>> vtk.inputs.voxeldims = [1,1,1]
    >>> vtk.run()                  # doctest: +SKIP
    """

    _cmd = "vtkstreamlines"
    input_spec = VtkStreamlinesInputSpec
    output_spec = VtkStreamlinesOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["vtk"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + ".vtk"


class ProcStreamlinesInputSpec(StdOutCommandLineInputSpec):
    inputmodel = traits.Enum(
        "raw",
        "voxels",
        argstr="-inputmodel %s",
        desc="input model type (raw or voxels)",
        usedefault=True,
    )

    in_file = File(
        exists=True,
        argstr="-inputfile %s",
        mandatory=True,
        position=1,
        desc="data file",
    )

    maxtractpoints = traits.Int(
        argstr="-maxtractpoints %d", units="NA", desc="maximum number of tract points"
    )
    mintractpoints = traits.Int(
        argstr="-mintractpoints %d", units="NA", desc="minimum number of tract points"
    )
    maxtractlength = traits.Int(
        argstr="-maxtractlength %d", units="mm", desc="maximum length of tracts"
    )
    mintractlength = traits.Int(
        argstr="-mintractlength %d", units="mm", desc="minimum length of tracts"
    )
    datadims = traits.List(
        traits.Int,
        desc="data dimensions in voxels",
        argstr="-datadims %s",
        minlen=3,
        maxlen=3,
        units="voxels",
    )
    voxeldims = traits.List(
        traits.Int,
        desc="voxel dimensions in mm",
        argstr="-voxeldims %s",
        minlen=3,
        maxlen=3,
        units="mm",
    )
    seedpointmm = traits.List(
        traits.Int,
        desc="The coordinates of a single seed point for tractography in mm",
        argstr="-seedpointmm %s",
        minlen=3,
        maxlen=3,
        units="mm",
    )
    seedpointvox = traits.List(
        traits.Int,
        desc="The coordinates of a single seed point for tractography in voxels",
        argstr="-seedpointvox %s",
        minlen=3,
        maxlen=3,
        units="voxels",
    )
    seedfile = File(
        exists=False, argstr="-seedfile %s", desc="Image Containing Seed Points"
    )
    regionindex = traits.Int(
        argstr="-regionindex %d", units="mm", desc="index of specific region to process"
    )
    iterations = traits.Float(
        argstr="-iterations %d",
        units="NA",
        desc="Number of streamlines generated for each seed. Not required when outputting streamlines, but needed to create PICo images. The default is 1 if the output is streamlines, and 5000 if the output is connection probability images.",
    )
    targetfile = File(
        exists=False, argstr="-targetfile %s", desc="Image containing target volumes."
    )
    allowmultitargets = traits.Bool(
        argstr="-allowmultitargets",
        desc="Allows streamlines to connect to multiple target volumes.",
    )
    directional = traits.List(
        traits.Int,
        desc="Splits the streamlines at the seed point and computes separate connection probabilities for each segment. Streamline segments are grouped according to their dot product with the vector (X, Y, Z). The ideal vector will be tangential to the streamline trajectory at the seed, such that the streamline projects from the seed along (X, Y, Z) and -(X, Y, Z). However, it is only necessary for the streamline trajectory to not be orthogonal to (X, Y, Z).",
        argstr="-directional %s",
        minlen=3,
        maxlen=3,
        units="NA",
    )
    waypointfile = File(
        exists=False,
        argstr="-waypointfile %s",
        desc="Image containing waypoints. Waypoints are defined as regions of the image with the same intensity, where 0 is background and any value > 0 is a waypoint.",
    )
    truncateloops = traits.Bool(
        argstr="-truncateloops",
        desc="This option allows streamlines to enter a waypoint exactly once. After the streamline leaves the waypoint, it is truncated upon a second entry to the waypoint.",
    )
    discardloops = traits.Bool(
        argstr="-discardloops",
        desc="This option allows streamlines to enter a waypoint exactly once. After the streamline leaves the waypoint, the entire streamline is discarded upon a second entry to the waypoint.",
    )
    exclusionfile = File(
        exists=False,
        argstr="-exclusionfile %s",
        desc="Image containing exclusion ROIs. This should be an Analyze 7.5 header / image file.hdr and file.img.",
    )
    truncateinexclusion = traits.Bool(
        argstr="-truncateinexclusion",
        desc="Retain segments of a streamline before entry to an exclusion ROI.",
    )

    endpointfile = File(
        exists=False,
        argstr="-endpointfile %s",
        desc="Image containing endpoint ROIs. This should be an Analyze 7.5 header / image file.hdr and file.img.",
    )

    resamplestepsize = traits.Float(
        argstr="-resamplestepsize %d",
        units="NA",
        desc="Each point on a streamline is tested for entry into target, exclusion or waypoint volumes. If the length between points on a tract is not much smaller than the voxel length, then streamlines may pass through part of a voxel without being counted. To avoid this, the program resamples streamlines such that the step size is one tenth of the smallest voxel dimension in the image. This increases the size of raw or oogl streamline output and incurs some performance penalty. The resample resolution can be controlled with this option or disabled altogether by passing a negative step size or by passing the -noresample option.",
    )

    noresample = traits.Bool(
        argstr="-noresample",
        desc="Disables resampling of input streamlines. Resampling is automatically disabled if the input model is voxels.",
    )

    outputtracts = traits.Bool(
        argstr="-outputtracts", desc="Output streamlines in raw binary format."
    )

    outputroot = File(
        exists=False,
        argstr="-outputroot %s",
        desc="Prepended onto all output file names.",
    )

    gzip = traits.Bool(argstr="-gzip", desc="save the output image in gzip format")
    outputcp = traits.Bool(
        argstr="-outputcp",
        desc="output the connection probability map (Analyze image, float)",
        requires=["outputroot", "seedfile"],
    )
    outputsc = traits.Bool(
        argstr="-outputsc",
        desc="output the connection probability map (raw streamlines, int)",
        requires=["outputroot", "seedfile"],
    )
    outputacm = traits.Bool(
        argstr="-outputacm",
        desc="output all tracts in a single connection probability map (Analyze image)",
        requires=["outputroot", "seedfile"],
    )
    outputcbs = traits.Bool(
        argstr="-outputcbs",
        desc="outputs connectivity-based segmentation maps; requires target outputfile",
        requires=["outputroot", "targetfile", "seedfile"],
    )


class ProcStreamlinesOutputSpec(TraitedSpec):
    proc = File(exists=True, desc="Processed Streamlines")
    outputroot_files = OutputMultiPath(File(exists=True))


class ProcStreamlines(StdOutCommandLine):
    """
    Process streamline data

    This program does post-processing of streamline output from track. It can either output streamlines or connection probability maps.
     * http://web4.cs.ucl.ac.uk/research/medic/camino/pmwiki/pmwiki.php?n=Man.procstreamlines

    Examples
    --------

    >>> import nipype.interfaces.camino as cmon
    >>> proc = cmon.ProcStreamlines()
    >>> proc.inputs.in_file = 'tract_data.Bfloat'
    >>> proc.run()                  # doctest: +SKIP
    """

    _cmd = "procstreamlines"
    input_spec = ProcStreamlinesInputSpec
    output_spec = ProcStreamlinesOutputSpec

    def _format_arg(self, name, spec, value):
        if name == "outputroot":
            return spec.argstr % self._get_actual_outputroot(value)
        return super(ProcStreamlines, self)._format_arg(name, spec, value)

    def __init__(self, *args, **kwargs):
        super(ProcStreamlines, self).__init__(*args, **kwargs)
        self.outputroot_files = []

    def _run_interface(self, runtime):
        outputroot = self.inputs.outputroot
        if isdefined(outputroot):
            actual_outputroot = self._get_actual_outputroot(outputroot)
            base, filename, ext = split_filename(actual_outputroot)
            if not os.path.exists(base):
                os.makedirs(base)
            new_runtime = super(ProcStreamlines, self)._run_interface(runtime)
            self.outputroot_files = glob.glob(
                os.path.join(os.getcwd(), actual_outputroot + "*")
            )
            return new_runtime
        else:
            new_runtime = super(ProcStreamlines, self)._run_interface(runtime)
            return new_runtime

    def _get_actual_outputroot(self, outputroot):
        actual_outputroot = os.path.join("procstream_outfiles", outputroot)
        return actual_outputroot

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["proc"] = os.path.abspath(self._gen_outfilename())
        outputs["outputroot_files"] = self.outputroot_files
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_proc"


class TractShredderInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True, argstr="< %s", mandatory=True, position=-2, desc="tract file"
    )

    offset = traits.Int(
        argstr="%d", units="NA", desc="initial offset of offset tracts", position=1
    )

    bunchsize = traits.Int(
        argstr="%d",
        units="NA",
        desc="reads and outputs a group of bunchsize tracts",
        position=2,
    )

    space = traits.Int(argstr="%d", units="NA", desc="skips space tracts", position=3)


class TractShredderOutputSpec(TraitedSpec):
    shredded = File(exists=True, desc="Shredded tract file")


class TractShredder(StdOutCommandLine):
    """
    Extracts bunches of streamlines.

    tractshredder works in a similar way to shredder, but processes streamlines instead of scalar data.
    The input is raw streamlines, in the format produced by track or procstreamlines.

    The program first makes an initial offset of offset tracts.  It then reads and outputs a group of
    bunchsize tracts, skips space tracts, and repeats until there is no more input.

    Examples
    --------

    >>> import nipype.interfaces.camino as cmon
    >>> shred = cmon.TractShredder()
    >>> shred.inputs.in_file = 'tract_data.Bfloat'
    >>> shred.inputs.offset = 0
    >>> shred.inputs.bunchsize = 1
    >>> shred.inputs.space = 2
    >>> shred.run()                  # doctest: +SKIP
    """

    _cmd = "tractshredder"
    input_spec = TractShredderInputSpec
    output_spec = TractShredderOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["shredded"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_shredded"


class DT2NIfTIInputSpec(CommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="-inputfile %s",
        mandatory=True,
        position=1,
        desc="tract file",
    )

    output_root = File(
        argstr="-outputroot %s",
        position=2,
        genfile=True,
        desc="filename root prepended onto the names of three output files.",
    )

    header_file = File(
        exists=True,
        argstr="-header %s",
        mandatory=True,
        position=3,
        desc=" A Nifti .nii or .hdr file containing the header information",
    )


class DT2NIfTIOutputSpec(TraitedSpec):
    dt = File(exists=True, desc="diffusion tensors in NIfTI format")

    exitcode = File(
        exists=True, desc="exit codes from Camino reconstruction in NIfTI format"
    )

    lns0 = File(
        exists=True, desc="estimated lns0 from Camino reconstruction in NIfTI format"
    )


class DT2NIfTI(CommandLine):
    """
    Converts camino tensor data to NIfTI format

    Reads Camino diffusion tensors, and converts them to NIFTI format as three .nii files.
    """

    _cmd = "dt2nii"
    input_spec = DT2NIfTIInputSpec
    output_spec = DT2NIfTIOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        output_root = self._gen_outputroot()
        outputs["dt"] = os.path.abspath(output_root + "dt.nii")
        outputs["exitcode"] = os.path.abspath(output_root + "exitcode.nii")
        outputs["lns0"] = os.path.abspath(output_root + "lns0.nii")
        return outputs

    def _gen_outfilename(self):
        return self._gen_outputroot()

    def _gen_outputroot(self):
        output_root = self.inputs.output_root
        if not isdefined(output_root):
            output_root = self._gen_filename("output_root")
        return output_root

    def _gen_filename(self, name):
        if name == "output_root":
            _, filename, _ = split_filename(self.inputs.in_file)
            filename = filename + "_"
        return filename


class NIfTIDT2CaminoInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="-inputfile %s",
        mandatory=True,
        position=1,
        desc="A NIFTI-1 dataset containing diffusion tensors. The tensors are assumed to be "
        "in lower-triangular order as specified by the NIFTI standard for the storage of "
        "symmetric matrices. This file should be either a .nii or a .hdr file.",
    )

    s0_file = File(
        argstr="-s0 %s",
        exists=True,
        desc="File containing the unweighted signal for each voxel, may be a raw binary "
        "file (specify type with -inputdatatype) or a supported image file.",
    )

    lns0_file = File(
        argstr="-lns0 %s",
        exists=True,
        desc="File containing the log of the unweighted signal for each voxel, may be a "
        "raw binary file (specify type with -inputdatatype) or a supported image file.",
    )

    bgmask = File(
        argstr="-bgmask %s",
        exists=True,
        desc="Binary valued brain / background segmentation, may be a raw binary file "
        "(specify type with -maskdatatype) or a supported image file.",
    )

    scaleslope = traits.Float(
        argstr="-scaleslope %s",
        desc="A value v in the diffusion tensor is scaled to v * s + i. This is "
        "applied after any scaling specified by the input image. Default is 1.0.",
    )

    scaleinter = traits.Float(
        argstr="-scaleinter %s",
        desc="A value v in the diffusion tensor is scaled to v * s + i. This is "
        "applied after any scaling specified by the input image. Default is 0.0.",
    )

    uppertriangular = traits.Bool(
        argstr="-uppertriangular %s",
        desc="Specifies input in upper-triangular (VTK style) order.",
    )


class NIfTIDT2CaminoOutputSpec(TraitedSpec):
    out_file = File(desc="diffusion tensors data in Camino format")


class NIfTIDT2Camino(CommandLine):
    """
    Converts NIFTI-1 diffusion tensors to Camino format. The program reads the
    NIFTI header but does not apply any spatial transformations to the data. The
    NIFTI intensity scaling parameters are applied.

    The output is the tensors in Camino voxel ordering: [exit, ln(S0), dxx, dxy,
    dxz, dyy, dyz, dzz].

    The exit code is set to 0 unless a background mask is supplied, in which case
    the code is 0 in brain voxels and -1 in background voxels.

    The value of ln(S0) in the output is taken from a file if one is supplied,
    otherwise it is set to 0.

    NOTE FOR FSL USERS - FSL's dtifit can output NIFTI tensors, but they are not
    stored in the usual way (which is using NIFTI_INTENT_SYMMATRIX). FSL's
    tensors follow the ITK / VTK "upper-triangular" convention, so you will need
    to use the -uppertriangular option to convert these correctly.

    """

    _cmd = "niftidt2camino"
    input_spec = NIfTIDT2CaminoInputSpec
    output_spec = NIfTIDT2CaminoOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self._gen_filename("out_file")
        return outputs

    def _gen_filename(self, name):
        if name == "out_file":
            _, filename, _ = split_filename(self.inputs.in_file)
        return filename


class AnalyzeHeaderInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=1,
        desc="Tensor-fitted data filename",
    )

    scheme_file = File(
        exists=True,
        argstr="%s",
        position=2,
        desc=("Camino scheme file (b values / vectors, " "see camino.fsl2scheme)"),
    )

    readheader = File(
        exists=True,
        argstr="-readheader %s",
        position=3,
        desc=(
            "Reads header information from file and prints to "
            "stdout. If this option is not specified, then the "
            "program writes a header based on the other "
            "arguments."
        ),
    )

    printimagedims = File(
        exists=True,
        argstr="-printimagedims %s",
        position=3,
        desc=(
            "Prints image data and voxel dimensions as " "Camino arguments and exits."
        ),
    )

    # How do we implement both file and enum (for the program) in one argument?
    # Is this option useful anyway?
    # -printprogargs <file> <prog>
    # Prints data dimension (and type, if relevant) arguments for a specific
    # Camino program, where prog is one of shredder, scanner2voxel,
    # vcthreshselect, pdview, track.
    printprogargs = File(
        exists=True,
        argstr="-printprogargs %s",
        position=3,
        desc=(
            "Prints data dimension (and type, if relevant) "
            "arguments for a specific Camino program, where "
            "prog is one of shredder, scanner2voxel, "
            "vcthreshselect, pdview, track."
        ),
    )

    printintelbyteorder = File(
        exists=True,
        argstr="-printintelbyteorder %s",
        position=3,
        desc=("Prints 1 if the header is little-endian, " "0 otherwise."),
    )

    printbigendian = File(
        exists=True,
        argstr="-printbigendian %s",
        position=3,
        desc=("Prints 1 if the header is big-endian, 0 " "otherwise."),
    )

    initfromheader = File(
        exists=True,
        argstr="-initfromheader %s",
        position=3,
        desc=(
            "Reads header information from file and "
            "intializes a new header with the values read "
            "from the file. You may replace any "
            "combination of fields in the new header by "
            "specifying subsequent options."
        ),
    )

    data_dims = traits.List(
        traits.Int,
        desc="data dimensions in voxels",
        argstr="-datadims %s",
        minlen=3,
        maxlen=3,
        units="voxels",
    )

    voxel_dims = traits.List(
        traits.Float,
        desc="voxel dimensions in mm",
        argstr="-voxeldims %s",
        minlen=3,
        maxlen=3,
        units="mm",
    )

    centre = traits.List(
        traits.Int,
        argstr="-centre %s",
        minlen=3,
        maxlen=3,
        units="mm",
        desc=(
            "Voxel specifying origin of Talairach "
            "coordinate system for SPM, default [0 0 0]."
        ),
    )

    picoseed = traits.List(
        traits.Int,
        argstr="-picoseed %s",
        minlen=3,
        maxlen=3,
        desc=("Voxel specifying the seed (for PICo maps), " "default [0 0 0]."),
        units="mm",
    )

    nimages = traits.Int(
        argstr="-nimages %d",
        units="NA",
        desc="Number of images in the img file. Default 1.",
    )

    datatype = traits.Enum(
        "byte",
        "char",
        "[u]short",
        "[u]int",
        "float",
        "complex",
        "double",
        argstr="-datatype %s",
        desc=(
            "The char datatype is 8 bit (not the 16 bit "
            "char of Java), as specified by the Analyze "
            "7.5 standard. The byte, ushort and uint "
            "types are not part of the Analyze "
            "specification but are supported by SPM."
        ),
        mandatory=True,
    )

    offset = traits.Int(
        argstr="-offset %d",
        units="NA",
        desc=(
            "According to the Analyze 7.5 standard, this is "
            "the byte offset in the .img file at which "
            "voxels start. This value can be negative to "
            "specify that the absolute value is applied for "
            "every image in the file."
        ),
    )

    greylevels = traits.List(
        traits.Int,
        argstr="-gl %s",
        minlen=2,
        maxlen=2,
        desc=("Minimum and maximum greylevels. Stored as " "shorts in the header."),
        units="NA",
    )

    scaleslope = traits.Float(
        argstr="-scaleslope %d",
        units="NA",
        desc=(
            "Intensities in the image are scaled by "
            "this factor by SPM and MRICro. Default is "
            "1.0."
        ),
    )

    scaleinter = traits.Float(
        argstr="-scaleinter %d",
        units="NA",
        desc=("Constant to add to the image intensities. " "Used by SPM and MRIcro."),
    )

    description = traits.String(
        argstr="-description %s",
        desc=(
            "Short description - No spaces, max "
            "length 79 bytes. Will be null "
            "terminated automatically."
        ),
    )

    intelbyteorder = traits.Bool(
        argstr="-intelbyteorder",
        desc=("Write header in intel byte order " "(little-endian)."),
    )

    networkbyteorder = traits.Bool(
        argstr="-networkbyteorder",
        desc=(
            "Write header in network byte order "
            "(big-endian). This is the default "
            "for new headers."
        ),
    )


class AnalyzeHeaderOutputSpec(TraitedSpec):
    header = File(exists=True, desc="Analyze header")


class AnalyzeHeader(StdOutCommandLine):
    """
    Create or read an Analyze 7.5 header file.

    Analyze image header, provides support for the most common header fields.
    Some fields, such as patient_id, are not currently supported. The program allows
    three nonstandard options: the field image_dimension.funused1 is the image scale.
    The intensity of each pixel in the associated .img file is (image value from file) * scale.
    Also, the origin of the Talairach coordinates (midline of the anterior commisure) are encoded
    in the field data_history.originator. These changes are included for compatibility with SPM.

    All headers written with this program are big endian by default.

    Example
    -------

    >>> import nipype.interfaces.camino as cmon
    >>> hdr = cmon.AnalyzeHeader()
    >>> hdr.inputs.in_file = 'tensor_fitted_data.Bdouble'
    >>> hdr.inputs.scheme_file = 'A.scheme'
    >>> hdr.inputs.data_dims = [256,256,256]
    >>> hdr.inputs.voxel_dims = [1,1,1]
    >>> hdr.run()                  # doctest: +SKIP
    """

    _cmd = "analyzeheader"
    input_spec = AnalyzeHeaderInputSpec
    output_spec = AnalyzeHeaderOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["header"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + ".hdr"


class ShredderInputSpec(StdOutCommandLineInputSpec):
    in_file = File(
        exists=True,
        argstr="< %s",
        mandatory=True,
        position=-2,
        desc="raw binary data file",
    )

    offset = traits.Int(
        argstr="%d", units="NA", desc="initial offset of offset bytes", position=1
    )

    chunksize = traits.Int(
        argstr="%d",
        units="NA",
        desc="reads and outputs a chunk of chunksize bytes",
        position=2,
    )

    space = traits.Int(argstr="%d", units="NA", desc="skips space bytes", position=3)


class ShredderOutputSpec(TraitedSpec):
    shredded = File(exists=True, desc="Shredded binary data file")


class Shredder(StdOutCommandLine):
    """
    Extracts periodic chunks from a data stream.

    Shredder makes an initial offset of offset bytes. It then reads and outputs
    chunksize bytes, skips space bytes, and repeats until there is no more input.

    If  the  chunksize  is  negative, chunks of size chunksize are read and the
    byte ordering of each chunk is reversed. The whole chunk will be reversed, so
    the chunk must be the same size as the data type, otherwise the order of the
    values in the chunk, as well as their endianness, will be reversed.

    Examples
    --------

    >>> import nipype.interfaces.camino as cam
    >>> shred = cam.Shredder()
    >>> shred.inputs.in_file = 'SubjectA.Bfloat'
    >>> shred.inputs.offset = 0
    >>> shred.inputs.chunksize = 1
    >>> shred.inputs.space = 2
    >>> shred.run()                  # doctest: +SKIP
    """

    _cmd = "shredder"
    input_spec = ShredderInputSpec
    output_spec = ShredderOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["shredded_file"] = os.path.abspath(self._gen_outfilename())
        return outputs

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_file)
        return name + "_shredded"



import pytest


from nipype.interfaces import dcm2nii


@pytest.mark.parametrize(
    "fname, extension",
    [
        ("output_1", ".txt"),
        ("output_w_[]_meta_1", ".json"),
        ("output_w_**^$?_meta_2", ".txt"),
    ],
)
def test_search_files(tmp_path, fname, extension):
    tmp_fname = fname + extension
    test_file = tmp_path / tmp_fname
    test_file.touch()
    actual_files_list = dcm2nii.search_files(str(tmp_path / fname), [extension])
    for f in actual_files_list:
        assert str(test_file) == f



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import Qwarp


def test_Qwarp_inputs():
    input_map = dict(
        Qfinal=dict(
            argstr="-Qfinal",
        ),
        Qonly=dict(
            argstr="-Qonly",
        ),
        allineate=dict(
            argstr="-allineate",
        ),
        allineate_opts=dict(
            argstr="-allineate_opts %s",
            requires=["allineate"],
        ),
        allsave=dict(
            argstr="-allsave",
            xor=["nopadWARP", "duplo", "plusminus"],
        ),
        args=dict(
            argstr="%s",
        ),
        ballopt=dict(
            argstr="-ballopt",
            xor=["workhard", "boxopt"],
        ),
        base_file=dict(
            argstr="-base %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        baxopt=dict(
            argstr="-boxopt",
            xor=["workhard", "ballopt"],
        ),
        blur=dict(
            argstr="-blur %s",
        ),
        duplo=dict(
            argstr="-duplo",
            xor=["gridlist", "maxlev", "inilev", "iniwarp", "plusminus", "allsave"],
        ),
        emask=dict(
            argstr="-emask %s",
            copyfile=False,
            extensions=None,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        expad=dict(
            argstr="-expad %d",
            xor=["nopadWARP"],
        ),
        gridlist=dict(
            argstr="-gridlist %s",
            copyfile=False,
            extensions=None,
            xor=["duplo", "plusminus"],
        ),
        hel=dict(
            argstr="-hel",
            xor=["nmi", "mi", "lpc", "lpa", "pear"],
        ),
        in_file=dict(
            argstr="-source %s",
            copyfile=False,
            extensions=None,
            mandatory=True,
        ),
        inilev=dict(
            argstr="-inilev %d",
            xor=["duplo"],
        ),
        iniwarp=dict(
            argstr="-iniwarp %s",
            xor=["duplo"],
        ),
        iwarp=dict(
            argstr="-iwarp",
            xor=["plusminus"],
        ),
        lpa=dict(
            argstr="-lpa",
            xor=["nmi", "mi", "lpc", "hel", "pear"],
        ),
        lpc=dict(
            argstr="-lpc",
            position=-2,
            xor=["nmi", "mi", "hel", "lpa", "pear"],
        ),
        maxlev=dict(
            argstr="-maxlev %d",
            position=-1,
            xor=["duplo"],
        ),
        mi=dict(
            argstr="-mi",
            xor=["mi", "hel", "lpc", "lpa", "pear"],
        ),
        minpatch=dict(
            argstr="-minpatch %d",
        ),
        nmi=dict(
            argstr="-nmi",
            xor=["nmi", "hel", "lpc", "lpa", "pear"],
        ),
        noXdis=dict(
            argstr="-noXdis",
        ),
        noYdis=dict(
            argstr="-noYdis",
        ),
        noZdis=dict(
            argstr="-noZdis",
        ),
        noneg=dict(
            argstr="-noneg",
        ),
        nopad=dict(
            argstr="-nopad",
        ),
        nopadWARP=dict(
            argstr="-nopadWARP",
            xor=["allsave", "expad"],
        ),
        nopenalty=dict(
            argstr="-nopenalty",
        ),
        nowarp=dict(
            argstr="-nowarp",
        ),
        noweight=dict(
            argstr="-noweight",
        ),
        num_threads=dict(
            nohash=True,
            usedefault=True,
        ),
        out_file=dict(
            argstr="-prefix %s",
            extensions=None,
            name_source=["in_file"],
            name_template="ppp_%s",
        ),
        out_weight_file=dict(
            argstr="-wtprefix %s",
            extensions=None,
        ),
        outputtype=dict(),
        overwrite=dict(
            argstr="-overwrite",
        ),
        pblur=dict(
            argstr="-pblur %s",
        ),
        pear=dict(
            argstr="-pear",
        ),
        penfac=dict(
            argstr="-penfac %f",
        ),
        plusminus=dict(
            argstr="-plusminus",
            xor=["duplo", "allsave", "iwarp"],
        ),
        quiet=dict(
            argstr="-quiet",
            xor=["verb"],
        ),
        resample=dict(
            argstr="-resample",
        ),
        verb=dict(
            argstr="-verb",
            xor=["quiet"],
        ),
        wball=dict(
            argstr="-wball %s",
            xor=["wmask"],
        ),
        weight=dict(
            argstr="-weight %s",
            extensions=None,
        ),
        wmask=dict(
            argstr="-wpass %s %f",
            xor=["wball"],
        ),
        workhard=dict(
            argstr="-workhard",
            xor=["boxopt", "ballopt"],
        ),
    )
    inputs = Qwarp.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Qwarp_outputs():
    output_map = dict(
        base_warp=dict(
            extensions=None,
        ),
        source_warp=dict(
            extensions=None,
        ),
        warped_base=dict(
            extensions=None,
        ),
        warped_source=dict(
            extensions=None,
        ),
        weights=dict(
            extensions=None,
        ),
    )
    outputs = Qwarp.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



""" This file contains defines parameters for nipy that we use to fill
settings in setup.py, the nipy top-level docstring, and for building the
docs.  In setup.py in particular, we exec this file, so it cannot import nipy
"""

# nipype version information
# Remove -dev for release
__version__ = "1.8.1"


def get_nipype_gitversion():
    """Nipype version as reported by the last commit in git

    Returns
    -------
    None or str
      Version of Nipype according to git.
    """
    import os
    import subprocess

    try:
        import nipype

        gitpath = os.path.realpath(
            os.path.join(os.path.dirname(nipype.__file__), os.path.pardir)
        )
    except:
        gitpath = os.getcwd()
    gitpathgit = os.path.join(gitpath, ".git")
    if not os.path.exists(gitpathgit):
        return None
    ver = None
    try:
        o, _ = subprocess.Popen(
            "git describe", shell=True, cwd=gitpath, stdout=subprocess.PIPE
        ).communicate()
    except Exception:
        pass
    else:
        ver = o.decode().strip().split("-")[-1]
    return ver


if __version__.endswith("-dev"):
    gitversion = get_nipype_gitversion()
    if gitversion:
        __version__ = "{}+{}".format(__version__, gitversion)

CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Environment :: Console",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: Apache Software License",
    "Operating System :: MacOS :: MacOS X",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Topic :: Scientific/Engineering",
]
PYTHON_REQUIRES = ">= 3.7"

description = "Neuroimaging in Python: Pipelines and Interfaces"

# Note: this long_description is actually a copy/paste from the top-level
# README.txt, so that it shows up nicely on PyPI.  So please remember to edit
# it only in one place and sync it correctly.
long_description = """========================================================
NIPYPE: Neuroimaging in Python: Pipelines and Interfaces
========================================================

Current neuroimaging software offer users an incredible opportunity to
analyze data using a variety of different algorithms. However, this has
resulted in a heterogeneous collection of specialized applications
without transparent interoperability or a uniform operating interface.

*Nipype*, an open-source, community-developed initiative under the
umbrella of `NiPy <http://nipy.org>`_, is a Python project that provides a
uniform interface to existing neuroimaging software and facilitates interaction
between these packages within a single workflow. Nipype provides an environment
that encourages interactive exploration of algorithms from different
packages (e.g., AFNI, ANTS, BRAINS, BrainSuite, Camino, FreeSurfer, FSL, MNE,
MRtrix, MNE, Nipy, Slicer, SPM), eases the design of workflows within and
between packages, and reduces the learning curve necessary to use different \
packages. Nipype is creating a collaborative platform for neuroimaging \
software development in a high-level language and addressing limitations of \
existing pipeline systems.

*Nipype* allows you to:

* easily interact with tools from different software packages
* combine processing steps from different software packages
* develop new workflows faster by reusing common steps from old ones
* process data faster by running it in parallel on many cores/machines
* make your research easily reproducible
* share your processing workflows with the community
"""

# versions
NIBABEL_MIN_VERSION = "2.1.0"
NETWORKX_MIN_VERSION = "2.0"
NUMPY_MIN_VERSION = "1.17"
SCIPY_MIN_VERSION = "0.14"
TRAITS_MIN_VERSION = "4.6"
DATEUTIL_MIN_VERSION = "2.2"
SIMPLEJSON_MIN_VERSION = "3.8.0"
PROV_MIN_VERSION = "1.5.2"
RDFLIB_MIN_VERSION = "5.0.0"
CLICK_MIN_VERSION = "6.6.0"
PYDOT_MIN_VERSION = "1.2.3"

NAME = "nipype"
MAINTAINER = "nipype developers"
MAINTAINER_EMAIL = "neuroimaging@python.org"
DESCRIPTION = description
LONG_DESCRIPTION = long_description
URL = "http://nipy.org/nipype"
DOWNLOAD_URL = "http://github.com/nipy/nipype/archives/master"
LICENSE = "Apache License, 2.0"
AUTHOR = "nipype developers"
AUTHOR_EMAIL = "neuroimaging@python.org"
PLATFORMS = "OS Independent"
MAJOR = __version__.split(".")[0]
MINOR = __version__.split(".")[1]
MICRO = __version__.replace("-", ".").split(".")[2]
ISRELEASE = (
    len(__version__.replace("-", ".").split(".")) == 3
    or "post" in __version__.replace("-", ".").split(".")[-1]
)
VERSION = __version__
PROVIDES = ["nipype"]
REQUIRES = [
    "click>=%s" % CLICK_MIN_VERSION,
    "networkx>=%s" % NETWORKX_MIN_VERSION,
    "nibabel>=%s" % NIBABEL_MIN_VERSION,
    "numpy>=%s" % NUMPY_MIN_VERSION,
    "packaging",
    "prov>=%s" % PROV_MIN_VERSION,
    "pydot>=%s" % PYDOT_MIN_VERSION,
    "python-dateutil>=%s" % DATEUTIL_MIN_VERSION,
    "rdflib>=%s" % RDFLIB_MIN_VERSION,
    "scipy>=%s" % SCIPY_MIN_VERSION,
    "simplejson>=%s" % SIMPLEJSON_MIN_VERSION,
    "traits>=%s,!=5.0" % TRAITS_MIN_VERSION,
    "filelock>=3.0.0",
    "etelemetry>=0.2.0",
    "looseversion",
]

TESTS_REQUIRES = [
    "codecov",
    "coverage<5",
    "pytest",
    "pytest-cov",
    "pytest-env",
    "pytest-timeout",
]

EXTRA_REQUIRES = {
    "data": ["datalad"],
    "doc": [
        "dipy",
        "ipython",
        "matplotlib",
        "nbsphinx",
        "sphinx-argparse",
        "sphinx>=2.1.2",
        "sphinxcontrib-apidoc",
    ],
    "duecredit": ["duecredit"],
    "nipy": ["nitime", "nilearn", "dipy", "nipy", "matplotlib"],
    "profiler": ["psutil>=5.0"],
    "pybids": ["pybids>=0.7.0"],
    "specs": ["black"],
    "ssh": ["paramiko"],
    "tests": TESTS_REQUIRES,
    "xvfbwrapper": ["xvfbwrapper"],
    # 'mesh': ['mayavi']  # Enable when it works
}


def _list_union(iterable):
    return list(set(sum(iterable, [])))


# Enable a handle to install all extra dependencies at once
EXTRA_REQUIRES["all"] = _list_union(EXTRA_REQUIRES.values())
# dev = doc + tests + specs
EXTRA_REQUIRES["dev"] = _list_union(
    val for key, val in EXTRA_REQUIRES.items() if key in ("doc", "tests", "specs")
)

STATUS = "stable"



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..preprocess import CAT12Segment


def test_CAT12Segment_inputs():
    input_map = dict(
        affine_preprocessing=dict(
            field="extopts.APP",
            usedefault=True,
        ),
        affine_regularization=dict(
            field="opts.affreg",
            usedefault=True,
        ),
        cobra=dict(
            field="output.ROImenu.atlases.hammers",
            usedefault=True,
        ),
        csf_output_dartel=dict(
            field="output.CSF.dartel",
            usedefault=True,
        ),
        csf_output_modulated=dict(
            field="output.CSF.mod",
            usedefault=True,
        ),
        csf_output_native=dict(
            field="output.CSF.native",
            usedefault=True,
        ),
        gm_output_dartel=dict(
            field="output.GM.dartel",
            usedefault=True,
        ),
        gm_output_modulated=dict(
            field="output.GM.mod",
            usedefault=True,
        ),
        gm_output_native=dict(
            field="output.GM.native",
            usedefault=True,
        ),
        hammers=dict(
            field="output.ROImenu.atlases.cobra",
            usedefault=True,
        ),
        ignore_errors=dict(
            field="extopts.ignoreErrors",
            usedefault=True,
        ),
        in_files=dict(
            copyfile=False,
            field="data",
            mandatory=True,
        ),
        initial_segmentation=dict(
            field="extopts.spm_kamap",
            usedefault=True,
        ),
        internal_resampling_process=dict(
            field="extopts.restypes.optimal",
            maxlen=2,
            minlen=2,
            usedefault=True,
        ),
        jacobianwarped=dict(
            field="output.jacobianwarped",
            usedefault=True,
        ),
        label_dartel=dict(
            field="output.label.dartel",
            usedefault=True,
        ),
        label_native=dict(
            field="output.label.native",
            usedefault=True,
        ),
        label_warped=dict(
            field="output.label.warped",
            usedefault=True,
        ),
        las_dartel=dict(
            field="output.las.dartel",
            usedefault=True,
        ),
        las_native=dict(
            field="output.las.native",
            usedefault=True,
        ),
        las_warped=dict(
            field="output.las.warped",
            usedefault=True,
        ),
        local_adaptive_seg=dict(
            field="extopts.LASstr",
            usedefault=True,
        ),
        lpba40=dict(
            field="output.ROImenu.atlases.lpba40",
            usedefault=True,
        ),
        matlab_cmd=dict(),
        mfile=dict(
            usedefault=True,
        ),
        n_jobs=dict(
            field="nproc",
            mandatory=True,
            usedefault=True,
        ),
        neuromorphometrics=dict(
            field="output.ROImenu.atlases.neuromorphometrics",
            usedefault=True,
        ),
        output_labelnative=dict(
            field="output.labelnative",
            usedefault=True,
        ),
        own_atlas=dict(
            copyfile=False,
            field="output.ROImenu.atlases.ownatlas",
            mandatory=False,
        ),
        paths=dict(),
        power_spm_inhomogeneity_correction=dict(
            field="opts.biasacc",
            usedefault=True,
        ),
        save_bias_corrected=dict(
            field="output.bias.warped",
            usedefault=True,
        ),
        shooting_tpm=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            field="extopts.registration.shooting.shootingtpm",
            mandatory=False,
        ),
        shooting_tpm_template_1=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            mandatory=False,
        ),
        shooting_tpm_template_2=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            mandatory=False,
        ),
        shooting_tpm_template_3=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            mandatory=False,
        ),
        shooting_tpm_template_4=dict(
            copyfile=False,
            extensions=[".hdr", ".img", ".img.gz", ".nii"],
            mandatory=False,
        ),
        skull_strip=dict(
            field="extopts.gcutstr",
            usedefault=True,
        ),
        surface_and_thickness_estimation=dict(
            field="surface",
            usedefault=True,
        ),
        surface_measures=dict(
            field="output.surf_measures",
            usedefault=True,
        ),
        tpm=dict(
            copyfile=False,
            field="tpm",
            mandatory=False,
        ),
        use_mcr=dict(),
        use_v8struct=dict(
            min_ver="8",
            usedefault=True,
        ),
        voxel_size=dict(
            field="extopts.vox",
            usedefault=True,
        ),
        warps=dict(
            field="output.warps",
            maxlen=2,
            minlen=2,
            usedefault=True,
        ),
        wm_hyper_intensity_correction=dict(
            field="extopts.WMHC",
            usedefault=True,
        ),
        wm_output_dartel=dict(
            field="output.WM.dartel",
            usedefault=True,
        ),
        wm_output_modulated=dict(
            field="output.WM.mod",
            usedefault=True,
        ),
        wm_output_native=dict(
            field="output.WM.native",
            usedefault=True,
        ),
    )
    inputs = CAT12Segment.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_CAT12Segment_outputs():
    output_map = dict(
        bias_corrected_image=dict(
            extensions=None,
        ),
        csf_dartel_image=dict(
            extensions=None,
        ),
        csf_modulated_image=dict(
            extensions=None,
        ),
        csf_native_image=dict(
            extensions=None,
        ),
        gm_dartel_image=dict(
            extensions=None,
        ),
        gm_modulated_image=dict(
            extensions=None,
        ),
        gm_native_image=dict(
            extensions=None,
        ),
        label_files=dict(),
        label_roi=dict(
            extensions=None,
        ),
        label_rois=dict(
            extensions=None,
        ),
        lh_central_surface=dict(
            extensions=None,
        ),
        lh_sphere_surface=dict(
            extensions=None,
        ),
        mri_images=dict(),
        report=dict(
            extensions=None,
        ),
        report_files=dict(),
        rh_central_surface=dict(
            extensions=None,
        ),
        rh_sphere_surface=dict(
            extensions=None,
        ),
        surface_files=dict(),
        wm_dartel_image=dict(
            extensions=None,
        ),
        wm_modulated_image=dict(
            extensions=None,
        ),
        wm_native_image=dict(
            extensions=None,
        ),
    )
    outputs = CAT12Segment.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..featuredetection import HammerAttributeCreator


def test_HammerAttributeCreator_inputs():
    input_map = dict(
        Scale=dict(
            argstr="--Scale %d",
        ),
        Strength=dict(
            argstr="--Strength %f",
        ),
        args=dict(
            argstr="%s",
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputCSFVolume=dict(
            argstr="--inputCSFVolume %s",
            extensions=None,
        ),
        inputGMVolume=dict(
            argstr="--inputGMVolume %s",
            extensions=None,
        ),
        inputWMVolume=dict(
            argstr="--inputWMVolume %s",
            extensions=None,
        ),
        outputVolumeBase=dict(
            argstr="--outputVolumeBase %s",
        ),
    )
    inputs = HammerAttributeCreator.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_HammerAttributeCreator_outputs():
    output_map = dict()
    outputs = HammerAttributeCreator.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..brainsuite import Cerebro


def test_Cerebro_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        costFunction=dict(
            argstr="-c %d",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        inputAtlasLabelFile=dict(
            argstr="--atlaslabels %s",
            extensions=None,
            mandatory=True,
        ),
        inputAtlasMRIFile=dict(
            argstr="--atlas %s",
            extensions=None,
            mandatory=True,
        ),
        inputBrainMaskFile=dict(
            argstr="-m %s",
            extensions=None,
        ),
        inputMRIFile=dict(
            argstr="-i %s",
            extensions=None,
            mandatory=True,
        ),
        keepTempFiles=dict(
            argstr="--keep",
        ),
        linearConvergence=dict(
            argstr="--linconv %f",
        ),
        outputAffineTransformFile=dict(
            argstr="--air %s",
            extensions=None,
            genfile=True,
        ),
        outputCerebrumMaskFile=dict(
            argstr="-o %s",
            extensions=None,
            genfile=True,
        ),
        outputLabelVolumeFile=dict(
            argstr="-l %s",
            extensions=None,
            genfile=True,
        ),
        outputWarpTransformFile=dict(
            argstr="--warp %s",
            extensions=None,
            genfile=True,
        ),
        tempDirectory=dict(
            argstr="--tempdir %s",
        ),
        tempDirectoryBase=dict(
            argstr="--tempdirbase %s",
        ),
        useCentroids=dict(
            argstr="--centroids",
        ),
        verbosity=dict(
            argstr="-v %d",
        ),
        warpConvergence=dict(
            argstr="--warpconv %f",
        ),
        warpLabel=dict(
            argstr="--warplevel %d",
        ),
    )
    inputs = Cerebro.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_Cerebro_outputs():
    output_map = dict(
        outputAffineTransformFile=dict(
            extensions=None,
        ),
        outputCerebrumMaskFile=dict(
            extensions=None,
        ),
        outputLabelVolumeFile=dict(
            extensions=None,
        ),
        outputWarpTransformFile=dict(
            extensions=None,
        ),
    )
    outputs = Cerebro.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""
Nipype base interfaces
----------------------

This module defines the API of all nipype interfaces.

"""
from traits.trait_handlers import TraitDictObject, TraitListObject
from traits.trait_errors import TraitError

from .core import (
    Interface,
    BaseInterface,
    SimpleInterface,
    CommandLine,
    StdOutCommandLine,
    MpiCommandLine,
    SEMLikeCommandLine,
    LibraryBaseInterface,
    PackageInfo,
)

from .specs import (
    BaseTraitedSpec,
    TraitedSpec,
    DynamicTraitedSpec,
    BaseInterfaceInputSpec,
    CommandLineInputSpec,
    StdOutCommandLineInputSpec,
)

from .traits_extension import (
    traits,
    Undefined,
    isdefined,
    has_metadata,
    File,
    ImageFile,
    Directory,
    Str,
    DictStrStr,
    OutputMultiObject,
    InputMultiObject,
    OutputMultiPath,
    InputMultiPath,
)

from .support import Bunch, InterfaceResult, NipypeInterfaceError



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..model import DualRegression


def test_DualRegression_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        con_file=dict(
            argstr="%s",
            extensions=None,
            position=4,
        ),
        des_norm=dict(
            argstr="%i",
            position=2,
            usedefault=True,
        ),
        design_file=dict(
            argstr="%s",
            extensions=None,
            position=3,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        group_IC_maps_4D=dict(
            argstr="%s",
            extensions=None,
            mandatory=True,
            position=1,
        ),
        in_files=dict(
            argstr="%s",
            mandatory=True,
            position=-1,
            sep=" ",
        ),
        n_perm=dict(
            argstr="%i",
            mandatory=True,
            position=5,
        ),
        one_sample_group_mean=dict(
            argstr="-1",
            position=3,
        ),
        out_dir=dict(
            argstr="%s",
            genfile=True,
            position=6,
            usedefault=True,
        ),
        output_type=dict(),
    )
    inputs = DualRegression.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DualRegression_outputs():
    output_map = dict(
        out_dir=dict(),
    )
    outputs = DualRegression.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value



# AUTO-GENERATED by tools/checkspecs.py - DO NOT EDIT
from ..registration import DiffeoSymTensor3DVol


def test_DiffeoSymTensor3DVol_inputs():
    input_map = dict(
        args=dict(
            argstr="%s",
        ),
        df=dict(
            argstr="-df %s",
            usedefault=True,
        ),
        environ=dict(
            nohash=True,
            usedefault=True,
        ),
        flip=dict(
            argstr="-flip %d %d %d",
        ),
        in_file=dict(
            argstr="-in %s",
            extensions=None,
            mandatory=True,
        ),
        interpolation=dict(
            argstr="-interp %s",
            usedefault=True,
        ),
        out_file=dict(
            argstr="-out %s",
            extensions=None,
            keep_extension=True,
            name_source="in_file",
            name_template="%s_diffeoxfmd",
        ),
        reorient=dict(
            argstr="-reorient %s",
            usedefault=True,
        ),
        resampling_type=dict(
            argstr="-type %s",
        ),
        target=dict(
            argstr="-target %s",
            extensions=None,
            xor=["voxel_size"],
        ),
        transform=dict(
            argstr="-trans %s",
            extensions=None,
            mandatory=True,
        ),
        voxel_size=dict(
            argstr="-vsize %g %g %g",
            xor=["target"],
        ),
    )
    inputs = DiffeoSymTensor3DVol.input_spec()

    for key, metadata in list(input_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(inputs.traits()[key], metakey) == value


def test_DiffeoSymTensor3DVol_outputs():
    output_map = dict(
        out_file=dict(
            extensions=None,
        ),
    )
    outputs = DiffeoSymTensor3DVol.output_spec()

    for key, metadata in list(output_map.items()):
        for metakey, value in list(metadata.items()):
            assert getattr(outputs.traits()[key], metakey) == value
